# Overrides compiler.transformer (which constructs AST nodes out of a
# parse tree). Adds parse tree elements to the relevant ast nodes This
# is useful because matching the scanner tokens to the parse tree is
# easier than matching them to the AST nodes.

#from extended_ast import *
from compiler.transformer import Transformer as TransformerBase
import compiler
import tokenutils
import tokenize
import token

import symbol
from compiler.consts import OP_ASSIGN, OP_DELETE, OP_APPLY


def parse(src):
    return Transformer().parse(src)

Base = TransformerBase
class Transformer(TransformerBase):
    
    def parsesuite(self, text):
        return TransformerBase.parsesuite(self, text)

    def pass_stmt(self, parsetree):
        n = Base.pass_stmt(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    def classdef(self, parsetree):
        n = Base.classdef(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    def funcdef(self, parsetree):
        n = Base.funcdef(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n


    def atom(self,parsetree):
        n = Base.atom(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    def import_stmt(self,parsetree):
        n = Base.import_stmt(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n
        

    def if_stmt(self, parsetree):
        n = Base.if_stmt(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    # handles Assign
    def expr_stmt(self, parsetree):
        n = Base.expr_stmt(self,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    # handles AssNode
    def com_assign_name(self, parsetree, assigning):
        n = Base.com_assign_name(self,parsetree,assigning)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    def addParsetreeToNodeIfItHasntAlreadyGotOne(self,node,parsetree):
        try:
            node.parsetree
        except AttributeError:
            node.parsetree = parsetree
            
    def addParsetreeToNodeOverridingExistingOne(self,node,parsetree):
        node.parsetree = parsetree


    # handles getAttr, callfunc
    def com_apply_trailer(self, primaryNode, parsetree):
        n = Base.com_apply_trailer(self, primaryNode, parsetree)
        parsetree = primaryNode.parsetree + parsetree
        #self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        self.addParsetreeToNodeOverridingExistingOne(n,parsetree)
        return n

    def print_stmt(self,parsetree):
        n = Base.print_stmt(self,parsetree)        
        if isinstance(n,compiler.ast.Print):
            n = Print(n.nodes,n.dest)
        elif isinstance(n,compiler.ast.Printnl):
            n = Printnl(n.nodes,n.dest)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n

    def com_binary(self, constructor, parsetree):
        n = Base.com_binary(self,constructor,parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def return_stmt(self, parsetree):
        n = Base.return_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def raise_stmt(self, parsetree):
        n = Base.raise_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def com_node(self, parsetree):
        n = Base.com_node(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    
    def not_test(self, parsetree):
        n = Base.not_test(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 
    
    def for_stmt(self, parsetree):
        n = Base.for_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 
        
    def break_stmt(self, parsetree):
        n = Base.break_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def continue_stmt(self, parsetree):
        n = Base.continue_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 


    def com_assign_tuple(self, parsetree, assigning):
        n = Base.com_assign_tuple(self, parsetree, assigning)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def com_assign_list(self, parsetree, assigning):
        n = Base.com_assign_list(self, parsetree, assigning)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 


    #def arith_expr(self, parsetree):
    #    n = Base.arith_expr(self, parsetree)
    #    print "brmnode",n
    #    self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
    #    return n 
    
    def com_try_except(self, parsetree):
        n = Base.com_try_except(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 
        
    def com_try_finally(self, parsetree):
        n = Base.com_try_finally(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def com_assign_attr(self, primary, parsetree, assigning):
        n = Base.com_assign_attr(self, primary, parsetree, assigning)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    
    def com_subscriptlist(self, primary, parsetree, assigning):
        #print "parsetreeeee",parsetree
        #print "primary",primary.parsetree
        #print "assigning",assigning
        #import inspect
        #print inspect.getouterframes(inspect.currentframe())
        #raise "arse"
        n = Base.com_subscriptlist(self, primary, parsetree, assigning)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    # handles Keyword
    def com_argument(self, parsetree, kw):
        v,n = Base.com_argument(self, parsetree, kw)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return v,n 


    
    def while_stmt(self, parsetree):
        n = Base.while_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def assert_stmt(self, parsetree):
        n = Base.assert_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def global_stmt(self, parsetree):
        n = Base.global_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    def exec_stmt(self, parsetree):
        n = Base.exec_stmt(self, parsetree)
        self.addParsetreeToNodeIfItHasntAlreadyGotOne(n,parsetree)
        return n 

    # Ideally we wouldn't need the next methods, but they create multiple AST nodes
    # internally, and so can't be 'decorated' like the above. Thus I've done a cut-n-paste job from 

    def arith_expr(self, parsetree):
        node = self.com_node(parsetree[0])
        for i in range(2, len(parsetree), 2):
            right = self.com_node(parsetree[i])
            leftparsetree = node.parsetree
            if parsetree[i-1][0] == token.PLUS:
                node = compiler.ast.Add([node, right])
                node.lineno = parsetree[1][2]
            elif parsetree[i-1][0] == token.MINUS:
                node = compiler.ast.Sub([node, right])
                node.lineno = parsetree[1][2]
            else:
                raise ValueError, "unexpected token: %s" % parsetree[i-1][0]
            node.parsetree = list(leftparsetree) + list(parsetree[i-1:i+1])
        return node


    def term(self, parsetree):
        #print "brm parsetree",parsetree[0]
        node = self.com_node(parsetree[0])
        #print "phil node",node, "parsetree", node.parsetree
        node.parsetree = parsetree[0]  # put the braces back on
        for i in range(2, len(parsetree), 2):
            #print "parsetree[i]",parsetree[i]
            right = self.com_node(parsetree[i])
            leftparsetree = node.parsetree   # <------ node already has a parsetree, and it's wrong
            #print "leftparsetree",leftparsetree
            t = parsetree[i-1][0]
            if t == token.STAR:
                node = compiler.ast.Mul([node, right])
            elif t == token.SLASH:
                node = compiler.ast.Div([node, right])
            elif t == token.PERCENT:
                node = compiler.ast.Mod([node, right])
            elif t == token.DOUBLESLASH:
                node = compiler.ast.FloorDiv([node, right])
            else:
                raise ValueError, "unexpected token: %s" % t
            node.lineno = parsetree[1][2]
            myparsetree = list(leftparsetree) + list(parsetree[i-1:i+1])
            # if brackets enclose the statement, gobble them
            #if (parsetree[i+2] == token.RBRACE):
            #    myparsetree = list(leftparsetree) + list(parsetree[i-1:i+1])
            self.addParsetreeToNodeIfItHasntAlreadyGotOne(node,myparsetree)
            #node.parsetree = list(leftparsetree) + list(parsetree[i-1:i+1])
        node.parsetree = parsetree
        return node


    def com_list_comprehension(self, expr, parsetree):
        lineno = parsetree[1][2]
        fors = []
        while parsetree:
            #print "parsetree",parsetree
            #print "parsetree[1]",parsetree[1]
            t = parsetree[1][1]
            if t == 'for':
                assignNode = self.com_assign(parsetree[2], OP_ASSIGN)
                listNode = self.com_node(parsetree[4])
                newfor = compiler.ast.ListCompFor(assignNode, listNode, [])
                newfor.lineno = parsetree[1][2]
                newfor.parsetree = parsetree[:5]
                fors.append(newfor)
                if len(parsetree) == 5:
                    parsetree = None
                else:
                    parsetree = self.com_list_iter(parsetree[5])
            elif t == 'if':
                test = self.com_node(parsetree[2])
                newif = compiler.ast.ListCompIf(test)
                newif.lineno = parsetree[1][2]
                newif.parsetree = parsetree[:3]
                newfor.ifs.append(newif)
                if len(parsetree) == 3:
                    parsetree = None
                else:
                    parsetree = self.com_list_iter(parsetree[3])
            else:
                raise SyntaxError, \
                      ("unexpected list comprehension element: %s %d"
                       % (node, lineno))
        n = compiler.ast.ListComp(expr, fors)
        n.lineno = lineno
        return n





    def parse(self,src):
        ast = self.parsesuite(src)
        return ast


# Print.getChildNodes() reverses the order of the nodes and
# destination, which buggers up the token matching. This class
# fixes that
class Printnl(compiler.ast.Printnl):

    def getChildNodes(self):
        nodes = []
        nodes.extend(compiler.ast.flatten_nodes(self.nodes))
        if self.dest is not None:
            nodes = [self.dest] + nodes
        return tuple(nodes)
        

# Print.getChildNodes() reverses the order of the nodes and
# destination, which buggers up the token matching. This class
# fixes that
class Print(compiler.ast.Print):

    def getChildNodes(self):
        nodes = []
        nodes.extend(compiler.ast.flatten_nodes(self.nodes))
        if self.dest is not None:
            nodes = [self.dest] + nodes
        return tuple(nodes)
