from __future__ import nested_scopes

import compiler
import tokenutils
import tokenize
import token
import brmtransformer

def matchTokensToASTFromSrc(ast, src):
    compiler.walk(ast, MatchTokensToAST.fromSrc(src))

def matchTokensToAST(ast, tokenlist):
    compiler.walk(ast, MatchTokensToAST(tokenlist))


from doublelinkedlist import region

class MatchTokensToAST:

    def fromSrc(src):
        tlist = tokenutils.getTokenList(src)
        return MatchTokensToAST(tlist)
    fromSrc = staticmethod(fromSrc)

    def __init__(self, tokenlist):
        self.tc = tokenutils.TokenConverter(tokenlist.begin(), tokenlist.end())

    def visitCompoundNode(self, node):
        # evaluate children first
        oldtokenconverter = self.tc
        self.tc = tokenutils.TokenConverter(oldtokenconverter.pos(), 
                                            oldtokenconverter.end())
        for child in node.getChildNodes():
            self.visit(child)
            # then evaluate compound node tokens
        self.tc = oldtokenconverter
        noderegion = self.tc.getTokensCorrespondingToParseTree(node.parsetree)
        self.addTokenIteratorsToASTNode(node, noderegion.begin, noderegion.end)
        #self.tc = tokenutils.TokenConverter(noderegion.begin,noderegion.end)        




    def visitModule(self, node):
        self.visit(node.node)
        self.addTokenIteratorsToASTNode(node, self.tc.begin(), self.tc.end())

    def visitStmt(self, node):
        for child in node.getChildNodes():
            self.visit(child)


    # non terminals
    visitDiscard = visitStmt
    visitCompare = visitStmt

    # special case
    def visitConst(self, node):
        if node.value != None:
            self.visitCompoundNode(node)

    def addTokenIteratorsToASTNode(self, node, begin, end):
        node.begin = begin
        node.end = end

    # all other nodes
    def __getattr__(self, name):
        if(name.startswith("visit")):
            return self.visitNode
        else:
            raise AttributeError, name

    def visitNode(self, node):
        self.visitCompoundNode(node)
        #for c in node.getChildNodes():
        #    self.visit(c,scope,sourcenode)
