import tokenize
from addtokens import AddTokens, addTokensToASTFromSrc
import tokenutils

import compiler

from tokenutils import TokenConverter
from brmtransformer import Transformer


# skip the newline, which is at the start of each testdata string
def skipNL(src):
    assert(src[0] == "\n")
    return src[1:]

def src2Tokens(src):
    return tokenutils.getTokenList(src)

class Mock:
    pass


def getAddTokens(src):
    ast = Transformer().parsesuite(src)
    tlist = tokenutils.getTokenList(src)
    addtokens = AddTokens(tlist)
    # need to get compiler to walk, since it adds a visit method to
    # target needed for testing
    compiler.walk(ast, addtokens)
    addtokens.__init__(tlist) # reinitialise the object
    return addtokens



class MockAddTokens(AddTokens):
    def __init__(self, tokenlist):
        AddTokens.__init__(self, "")
        self.tc = TokenConverter(tokenlist)

    # will be called to handle body of function, class ..etc..
    def visit(self, code):
        self.removePassTokensFromTokenList()

    def removePassTokensFromTokenList(self):
        assert(self.tc.popToken().name == "pass")
        assert(self.tc.popToken().type == tokenize.NEWLINE)


class MockAddTokens_codeIsInlined(MockAddTokens):
    def removePassTokensFromTokenList(self):
        assert(self.tc.popToken().name == "pass")
        # no newline expected, because code is inlined


# common code for testing compound nodes (e.g. classes, functions)
class CompoundNodeTestHelper:

    def check_CompoundNode_GetsAllTokensExceptPassToken(self, node, src):
        self.assertGetTokensBefore_returnsAllTokensUpToPassToken(node, src)
        self.assertGetTokensAfter_returnsDedentAfterPassToken(node, src)

    def assertGetTokensBefore_returnsAllTokensUpToPassToken(self, node, src):
        assert(node.getTokensBefore() == src2Tokens(src)[:-4])

    def assertGetTokensAfter_returnsDedentAfterPassToken(self, node, src):
        dedent = src2Tokens(src)[-2]
        assert(dedent.type == tokenize.DEDENT)
        assert(node.getTokensAfter() == [dedent])
