#!/usr/bin/env python 
from vector import Vector
from doublelinkedlist import DoubleLinkedList
import unittest

class TestVector(unittest.TestCase):

    def getContainer(self):
        return Vector()

    def _assertNoMoreElements(self,iteratorOrList):
        it = iter(iteratorOrList)
        try:
            it.next()
            self.fail("shouldnt get to here")
        except StopIteration: pass
        
    def test_iterationReturnsNoElementsForEmptyList(self):
        l = self.getContainer()
        self._assertNoMoreElements(l)        

    def test_beginEqualsEndForEmptyList(self):
        l = self.getContainer()
        assert l.begin() == l.end()


    def test_preincrReturnsElementsInRightOrder(self):
        l = self.getContainer()
        l.append(1)
        l.append(2)
        l.append(3)
        it = l.begin()
        self.assertEqual(it.deref(), 1)
        self.assertEqual(it.preincr().deref(),2)
        self.assertEqual(it.preincr().deref(),3)


    def test_iterationReturnsElementsInRightOrder(self):
        l = self.getContainer()
        l.append(1)
        l.append(2)
        it = iter(l)
        self.assertEqual(it.next(),1)
        self.assertEqual(it.next(),2)
        self._assertNoMoreElements(it)

    def test_insertAtBeginningWorksWithEmptyList(self):
        l = self.getContainer()
        it = iter(l)
        l.insert(it,0)
        it = iter(l)
        self.assertEqual(it.next(),0)
        self._assertNoMoreElements(it)

    def test_insertAtBeginningWorksWithFullList(self):
        l = self.getContainer()
        l.append(1)
        l.append(2)
        it = iter(l)
        l.insert(it,0)
        it = iter(l)
        self.assertEqual(it.next(),0)
        self.assertEqual(it.next(),1)
        self.assertEqual(it.next(),2)
        self._assertNoMoreElements(it)

    def test_insertinMiddleWorks(self):
        l = self.getContainer()
        l.append(0)
        l.append(2)
        it = iter(l)
        it.preincr()
        l.insert(it,1)
        it = iter(l)
        self.assertEqual(it.deref(),0)
        self.assertEqual(it.preincr().deref(),1)
        self.assertEqual(it.preincr().deref(),2)
        assert it.preincr() == l.end()
        self.assertEqual(it.predecr().deref(),2)
        self.assertEqual(it.predecr().deref(),1)
        self.assertEqual(it.predecr().deref(),0)
        assert it == l.begin()
        

    def test_predecrReturnsElementsInRightOrder(self):
        l = self.getContainer()
        l.append(0)
        l.append(1)
        l.append(2)
        it = iter(l)
        it.preincr()
        it.preincr()
        self.assertEqual(it.predecr().deref(),1)
        self.assertEqual(it.predecr().deref(),0)
        assert l.begin() == it

    def test_cloneReturnesCopiedIterator(self):
        l = self.getContainer()
        l.append(0)
        l.append(1)
        l.append(2)
        it = iter(l)
        self.assertEqual(it.preincr().deref(),1)
        it2 = it.clone()
        self.assertEqual(it2.deref(),1)
        self.assertEqual(it.preincr().deref(),2)
        self.assertEqual(it2.preincr().deref(),2)

    def test_deleteRemovesItem(self):
        l = self.getContainer()
        l.append(0)
        l.append(1)
        l.append(2)
        it = iter(l)
        it.preincr()
        l.delete(it)
        it = iter(l)
        self.assertEqual(it.deref(),0)
        self.assertEqual(it.preincr().deref(),2)
        self._assertNoMoreElements(it.preincr())

    def test_replaceReplacesItem(self):
        l = self.getContainer()
        l.append(0)
        l.append(1)
        l.append(2)
        it = l.begin().clone()
        it.preincr()
        l.replace(it,3)
        self.assertEqual(it.deref(),3)
        it = l.begin().clone()
        self.assertEqual(it.deref(),0)
        self.assertEqual(it.preincr().deref(),3)
        self.assertEqual(it.preincr().deref(),2)
        self._assertNoMoreElements(it.preincr())


class TestDoubleLinkedList(TestVector):
    def getContainer(self):
        return DoubleLinkedList()
    
    """
    def test_deleteRegion(self):
        l = self.getContainer()
        l.append(0)
        l.append(1)
        l.append(2)
        l.append(3)
        startit = l.begin().clone()
        startit.preincr()
        endit = startit.clone()
        endit.preincr()
        endit.preincr()
        l.deleteRegion(startit,endit)
        it = l.begin().clone()
        self.assertEqual(it.deref(),0)
        self.assertEqual(it.preincr().deref(),3)
        self._assertNoMoreElements(it.preincr())

        

    def test_insertRegionInsertsRegion(self):
        l = self.getContainer()
        l.append(0)
        l.append(3)
        l.append(4)
        l.append(1)
        l.append(2)
        startit = l.begin()
        startit.preincr()
        endit = startit.clone()
        endit.preincr()
        endit.preincr()
        it = l.begin().clone()
        l.deleteRegion(startit,endit)
        print startit.deref(),endit.deref()
        
        print l
        l.insertRegion(it,startit,endit)
        print l
        it = l.begin().clone()
        self.assertEqual(it.deref(),0)
        self.assertEqual(it.preincr().deref(),1)
        self.assertEqual(it.preincr().deref(),2)
        self.assertEqual(it.preincr().deref(),3)
        self.assertEqual(it.preincr().deref(),4)
        self._assertNoMoreElements(it.preincr())
    """


    def test_moveRegionMovesRegionWhenRegionAtStart(self):
        l = self.getContainer()
        l.append(2)
        l.append(3)
        l.append(1)
        l.append(4)
        startit = l.begin()
        endit = startit.clone()
        endit.preincr().preincr()
        it = endit.clone()
        it.preincr()
        l.moveRegion(startit,endit,it)
        it = l.begin().clone()
        self.assertEqual(it.deref(),1)
        self.assertEqual(it.preincr().deref(),2)
        self.assertEqual(it.preincr().deref(),3)
        self.assertEqual(it.preincr().deref(),4)


    def test_moveRegionMovesRegionWhenRegionAtEnd(self):
        l = self.getContainer()
        l.append(3)
        l.append(4)
        l.append(1)
        l.append(2)
        startit = l.begin().clone()
        startit.preincr().preincr()
        endit = startit.clone()
        endit.preincr().preincr()
        it = l.begin().clone()
        l.moveRegion(startit,endit,it)
        it = l.begin().clone()
        self.assertEqual(it.deref(),1)
        self.assertEqual(it.preincr().deref(),2)
        self.assertEqual(it.preincr().deref(),3)
        self.assertEqual(it.preincr().deref(),4)

    def test_moveRegionMovesRegionWhenRegionInMiddle(self):
        l = self.getContainer()
        l.append(1)
        l.append(3)
        l.append(4)
        l.append(2)
        l.append(5)
        startit = l.begin().clone()
        startit.preincr()
        endit = startit.clone()
        endit.preincr().preincr()
        it = endit.clone()
        it.preincr()
        l.moveRegion(startit,endit,it)
        it = l.begin().clone()
        self.assertEqual(it.deref(),1)
        self.assertEqual(it.preincr().deref(),2)
        self.assertEqual(it.preincr().deref(),3)
        self.assertEqual(it.preincr().deref(),4)
        self.assertEqual(it.preincr().deref(),5)
        
    def test_insertRegionWorksAtStartOfRegion(self):
        l = self.getContainer()
        l = self.getContainer()
        l.append(0)
        
if __name__ == "__main__":
    unittest.main()
