#!/usr/bin/env python 
import setpath

import unittest

from tokenutils import getTokenList, getTokenVector

#from test_addtokens_support import*

#from test_addtokens_support import MockAddTokens_codeIsInlined

import brmtransformer
from brmtransformer import parse

from bike.testutils import*

from bike import testdata
from doublelinkedlist import region

def skipNL(src):
    assert(src[0] == "\n")
    return src[1:]

def assertTokensAreSame(iterators, tokens):
    begin,end = iterators
    it = begin.clone()
    pos = 0
    while it != end:
        assert it.deref() == tokens[pos]
        it.preincr()
        pos += 1
    assert len(tokens) == pos


class TestVisitClass(unittest.TestCase):

    def test_getsClassTokens(self):
        src = skipNL(testdata.class_and_method)
        fn = getASTNodeFromSrc(src,"Class")
        begin,end = fn.getTokenIterators()
        toks = getTokenVector(src)
        assertTokensAreSame((begin,end),toks[:-1])

class TestVisitFunction(unittest.TestCase):

    def test_getsFunctionTokens(self):
        src = skipNL(testdata.BasicFunction)
        fn = getASTNodeFromSrc(src,"Function")
        begin,end = fn.getTokenIterators()
        toks = getTokenVector(src)

        assertTokensAreSame((begin,end),toks[:-1])

    def test_doesntIncludePreceedingComments(self):
        src = skipNL(testdata.function_withPreceedingComment)
        fn = getASTNodeFromSrc(src,"Function")
        begin,end = fn.getTokenIterators()
        toks = getTokenVector(src)
        assertTokensAreSame((begin,end),toks[2:-1])


class TestVisitGetattr(unittest.TestCase):
    
    def test_getsGetattrTokens(self):
        src = skipNL(testdata.FunctionWithGetattr)
        fn = getASTNodeFromSrc(src,"Getattr")
        begin,end = fn.getTokenIterators()
        toks = getTokenVector(src)
        assertTokensAreSame((begin,end),toks[7:10])


class TestVisitAdd(unittest.TestCase):

    def test_getsCorrectTokensForSimpleAdd(self):
        src = "1 + 2"
        a = getASTNodeFromSrc(src,"Add")
        begin,end = a.getTokenIterators()
        toks = getTokenVector(src)
        assertTokensAreSame((begin,end),toks[:-1])
        
    def test_getsCorrectTokensForCompoundAdd(self):
        src = "1 + 2 + 3 + 4"
        a = getASTNodeFromSrc(src,"Add",0)
        assertTokensMatch(a,getTokenVector(src)[:-1])
        a = getASTNodeFromSrc(src,"Add",1)
        assertTokensMatch(a,getTokenVector(src)[:-3])
        a = getASTNodeFromSrc(src,"Add",2)
        assertTokensMatch(a,getTokenVector(src)[:-5])

    def test_getsCorrectTokensForCompoundSubtract(self):
        src = "1 - 2 - 3 - 4"
        a = getASTNodeFromSrc(src,"Sub",0)
        assertTokensMatch(a,getTokenVector(src)[:-1])
        a = getASTNodeFromSrc(src,"Sub",1)
        assertTokensMatch(a,getTokenVector(src)[:-3])
        a = getASTNodeFromSrc(src,"Sub",2)
        assertTokensMatch(a,getTokenVector(src)[:-5])


class TestVisitMul(unittest.TestCase):
    def test_getsCorrectTokensForCompoundMul(self):
        src = "1 * 2 * 3 * 4"
        a = getASTNodeFromSrc(src,"Mul",0)
        assertTokensMatch(a,getTokenVector(src)[:-1])
        a = getASTNodeFromSrc(src,"Mul",1)
        assertTokensMatch(a,getTokenVector(src)[:-3])
        a = getASTNodeFromSrc(src,"Mul",2)
        assertTokensMatch(a,getTokenVector(src)[:-5])


def assertTokensMatch(ast,tokens):
    begin,end = ast.getTokenIterators()
    try:
        assertTokensAreSame((begin,end),tokens)
    except AssertionError:
        raise AssertionError(str(region(begin,end))+" != "+str(tokens))
    
class TestVisitListComp(unittest.TestCase):
    def test_getsCorrectTokensForCompoundFor(self):
        src = "[x.name for x in [y.name for y in self.packages]]"
        toks = getTokenVector(src)
        node = getASTNodeFromSrc(src,"ListComp",0)
        assertTokensMatch(node,toks[:-1])
        node = getASTNodeFromSrc(src,"ListCompFor")
        assertTokensMatch(node,toks[4:-2])
        node = getASTNodeFromSrc(src,"ListComp",1)
        assertTokensMatch(node,toks[7:-2])
        node = getASTNodeFromSrc(src,"ListCompFor",1)
        assertTokensMatch(node,toks[11:-3])

    def test_getsCorrectTokensForCompoundForIf(self):
        src = "[x.name for x in a if x.foo == 3]"
        toks = getTokenVector(src)
        node = getASTNodeFromSrc(src,"ListComp",0)
        assertTokensMatch(node,toks[:-1])
        node = getASTNodeFromSrc(src,"ListCompFor")
        assertTokensMatch(node,toks[4:-8])
        node = getASTNodeFromSrc(src,"ListCompIf")
        assertTokensMatch(node,toks[8:-2])


        

class TestHandlesAllTokens(unittest.TestCase):
    def test_handlesAllTokens(self):
        src = skipNL(testdata.source_with_all_ast_nodes)        
        #src = readFile("./scrap.py")
        #src = skipNL(data)
        #import parser
        #print parser.suite(src).totuple()

        node = getASTNodeFromSrc(src,"Module")
        #print node
        #print node.parsetree
        begin,end = node.getTokenIterators()

data = """
(2 + 3)+hue+6
"""
if __name__ == "__main__":
    unittest.main()
