#!/usr/bin/env python 
import setpath
import unittest
import os
from bike.parsing.extended_ast import Top, Source

from test_load import tmpfile, writeTmpTestFile

from bike import testdata
from tokenutils import getTokenList
from bike.mock import Mock
from brmtransformer import parse
from output import*
from matchTokensToAST import matchTokensToASTFromSrc
from bike.testutils import*



class TestTokensToString(unittest.TestCase):
    def testTokensToString_withSource(self):
        src = testdata.source_with_all_ast_nodes[1:]
        ast = createAST(src)
        newsrc = ast.getSourceAsString()
        #print newsrc
        #print "-----------------"
        #print src
        assert newsrc == src

    def testTokensToString_withClass(self):
        src = testdata_class[1:]
        ast = createAST(src)
        assert(src == ast.getSourceAsString())

    def testTokensToString_withClassAndMethod(self):
        src = testdata.class_and_method[1:]
        ast = createAST(src)
        out = ast.getSourceAsString()
        assert(src == out)

    # I like my code to be formatted  a = b, not a=b
    def testTokensToString_addsASpaceWhenHandlingAssignments(self):
        src = "a = b"
        ast = createAST(src)
        assert ast.getSourceAsString() == src

    # a.b, not a. b
    def testTokensToString_putsNoSpacesInGetAttrDot(self):
        src = "a.b"
        ast = createAST(src)
        assert ast.getSourceAsString() == src

    # a == b, not a==b
    def testTokensToString_addsSpacesWhenHandlingConditionals(self):
        for src in "a == b", "a != b", "a < b", "a <= b", "a > b", "a >= b", "a in b":
            ast = createAST(src)
            assert ast.getSourceAsString() == src, "adds space when handling "+src

    def testTokensToString_addsSpaceAfterColonWhenDoingInlineBlocks(self):
        src = "def foo(): pass"
        ast = createAST(src)
        assert ast.getSourceAsString() == src

    def testTokensToString_addsTwoSpacesWhenDoingInlineComments(self):
        src = "pass  # comment"
        ast = createAST(src)
        assert ast.getSourceAsString() == src

    def testTokensToString_addsASpaceAfterEachComma(self):
        src = testdata.StreamToStderr[1:]
        ast = createAST(src)
        assert ast.getSourceAsString() == src

    def testTokensToString_handlesLineBreaks(self):
        src = testdata.linebreak[1:]
        ast = createAST(src)
        assert ast.getSourceAsString() == src

    def testTokensToString_handlesLineBreaksIfTokenHasBeenMoved(self):
        src = testdata.linebreak[1:]
        ast = createAST(src)
        tokens = ast.getTokens()
        it = tokens.begin().clone()
        # modify the token
        for i in range(12): it.preincr()
        from bike.parsing.tokenutils import TokenList
        tokens.replace(it, TokenList.token(tokenize.NAME, 'babble', 
                                          (3, 1), (0, 0), "babble"))
        assert tokensToString(tokens.begin(), tokens.end()) == testdata.linebreak_outputIfTokensMoved[1:]

    def testTokensToString_returnsEmptyStringIfGivenEmptyList(self):
        tokens = DoubleLinkedList()
        assert tokensToString(tokens.begin(), tokens.end()) == ""


testdata_class = """
class foo:
    pass
"""



class TestCommentIncludesNewline(unittest.TestCase):

    def testCommentIncludesNewline_returnsTrueIfCommentEndsInNewline(self):
        src = testdata_comment_ending_in_newline[1:]
        tokens = getTokenList(src)
        assert TokensToString().commentIncludesNewline(tokens.begin().deref())

    def testCommentIncludesNewline_returnsFalseIfDoesntContainNewline(self):
        src = testdata_comment_not_ending_in_newline[1:]
        tokenit = getTokenList(src).begin()
        assert TokensToString().commentIncludesNewline(tokenit.preincr().deref()) == 0



testdata_comment_ending_in_newline = """
# comment
"""

testdata_comment_not_ending_in_newline = """
pass # comment
"""


class TestSave(unittest.TestCase):
    def testSave_doesntWriteFilesIfSourceNotModified(self):
        writeTmpTestFile("notpass")
        s = Source(tmpfile, "foobahfile", "pass")
        xast = Top([], [s], "")
        save(xast)  # setModified() not called on source!
        open(tmpfile).read() == "notpass"  # file hasnt changed
        deleteTmpTestFile()


    def testSave_writesFilesIfSourceModified(self):
        try:
            writeTmpTestFile("notpass")
            s = Source(tmpfile, "foobahfile", "pass")
            xast = Top([], [s], "")
            s.tokenize()
            s.setModified()
            filesWritten = save(xast)
            assert open(tmpfile).read() == "pass"
            assert filesWritten == [os.path.abspath(tmpfile)]
        finally:
            deleteTmpTestFile()




if __name__ == "__main__":
    unittest.main()
