#!/usr/bin/env python
import setpath
import unittest
import parser
import symbol
import token
import tokenize

import tokenutils

from tokenutils import getTokenList
from tokenutils import TokenConverter
from tokenutils import getTokenConverter
import testdata
from bike import testdata as globaltestdata
from doublelinkedlist import region, DoubleLinkedList

class TestTokenEquality(unittest.TestCase):
    #test data
    token =  tokenutils.TokenList.token # less typing below:
    tok = token(tokenize.NAME,"hello",(0,0),(0,0),"hello")
    tok_dup = token(tokenize.NAME,"hello",(0,0),(0,0),"hello")
    not_tok = [
        token(tokenize.OP,"hello",(0,0),(0,0),"hello"),
        token(tokenize.NAME,"hell",(0,0),(0,0),"hello"),
        ]
    

    def test__eq__(self):
        assert(TestTokenEquality.tok == TestTokenEquality.tok_dup)
        for i in TestTokenEquality.not_tok:            
            assert (TestTokenEquality.tok != i),"failed with "+str(i)

    def test__ne__(self):
        assert((TestTokenEquality.tok != TestTokenEquality.tok_dup) == 0) 



class TestTokenList(unittest.TestCase):
    def testTokenList(self):
        # basic test to make sure that list is produced
        tokenlist = getTokenList("class foo")
        it = tokenlist.begin()
        assert(it.deref().name == "class")
        assert(it.deref().type == tokenize.NAME)
        it.preincr()
        assert(it.deref().name == "foo")
        assert(it.deref().type == tokenize.NAME)
        it.preincr()
        assert(it.deref().type == tokenize.ENDMARKER)
        assert(it.preincr() == tokenlist.end())


    # test to check that comments immediately before a block of code
    # are placed after the previous block DEDENT token
    def testTokenList_commentsBeforeCodeBlockOccurAfterPreviousDedentToken(self):
        t = getTokenList(testdata_reorderComment_class)
        # find dedent
        it = iter(t)
        while it.deref().type != token.DEDENT:
            it.preincr()

        # check that dedent comes after comment describing prev fn.
        it2 = it.clone()
        assert it2.predecr().deref().name == '# previous comment\n'

        # and before the newline and comment describing next class
        assert(it.preincr().deref().type == tokenize.NL)
        assert(it.preincr().deref().name == '# This comment should come after the previous class dedent. It\n')
        




    def testTokenList_commentsDescribingClassMemberOccurAfterIndent(self):
        it = getTokenList(testdata_classWithCommentOnNextLine).begin()
        while it.deref().type != token.INDENT: it.preincr()
        # assert comment comes after indent
        assert it.preincr().deref().type == tokenize.COMMENT
        
        

testdata_classWithCommentOnNextLine = """
class foo:
    # hello
    pass
"""

testdata_reorderComment_class = """
class foo:
    pass  # this comment should come before class dedent
          # so should this - it might be a continuation of the
          # previous comment

# This comment should come after the previous class dedent. It
# probably documents the next class or group of classes

# so should this
class bah:
    pass
"""

testdata_reorderComment_function = """
class foo:

    def a(self):
        pass  # this comment should come before function dedent
              # so should this - it's probably be a continuation of
              # the previous comment

    # This comment should come after the previous function dedent. It
    # probably documents the next class or group of classes

    # so should this
    def b(self):
        pass
    
"""


class TestNodeIsToken(unittest.TestCase):

    tokenNode = [tokenize.NAME, 'class', 1]
    notTokenNode = [250, '', 1]  

    def testNodeIsToken_returnsTrueForToken(self):
        l = DoubleLinkedList()
        tc = TokenConverter(l.begin(),l.end())
        assert(tc._nodeIsToken(TestNodeIsToken.tokenNode))

    def testNodeIsToken_returnsFalseForNonToken(self):
        l = DoubleLinkedList()
        tc = TokenConverter(l.begin(),l.end())
        assert(tc._nodeIsToken(TestNodeIsToken.notTokenNode) == 0)


def getParseTree(src):
    return parser.suite(src).tolist(1)



class TestGetTokensCorrespondingToParseTree(unittest.TestCase):
    # tests removed because they are no longer relevant
    # functionality at tested higher level
    """
    def test_returnsCorrectTokenListWhenTokensAreAtEnd(self):
        src = testdata_getAttrTwice[1:]
        parsetree = getParseTree(src)
        tlist = getTokenList(src)
        tc = TokenConverter(tlist.begin(),tlist.end())
        begin = tlist.begin().clone()
        end = tlist.end().clone()
        for i in range(4): begin.preincr()
        for i in range(2): end.predecr()
        #print parsetree[2][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1]
        print tc.getTokensCorrespondingToParseTree(parsetree)
        print region(begin,end)
        #self.assertEqual(tc.getTokensCorrespondingToParseTree(parsetree[2]),
        #                 region(begin,end))
    """
testdata_getAttrTwice = """
a.b
a.c
"""

if __name__ == "__main__":
    unittest.main()
    #import pdb
    #pdb.run('unittest.main()')
