"""
This module contains base classes which hold and manipulate token
state for ast Nodes.
The ast nodes subclass these classes by means of hackery in the extended_ast.py module

Author Phil Dawes (pdawes@users.sourceforge.net)
"""

from compiler.ast import Node
import tokenize
from utils import fqn_rcdr, fqn_rcar
from doublelinkedlist import DoubleLinkedList, region

class coords:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return "("+str(self.x)+","+str(self.y)+")"

class SimpleTokenHandler:
    def _getTokens(self):
        if not hasattr(self, "sourcenode"):
            print self

        # sourcenode is added to ast nodes by addtypeinfo
        assert hasattr(self, "sourcenode")
        self.sourcenode.tokenize()
        return self.getTokens_impl()

    def getTokenIterators(self):
        if not hasattr(self, "sourcenode"):
            print self
        # sourcenode is added to ast nodes by addtypeinfo
        assert hasattr(self, "sourcenode")
        self.sourcenode.tokenize()
        return self.begin, self.end

    def getNodeTokenIterators(self):
        return self.getTokenIterators()

    def addTokenIterators(self, tokenIterators):
        self.begin,self.end = tokenIterators
        assert self.begin != self.end

    def getTokens_impl(self):
        if hasattr(self, "tokens"):
            return self.tokens
        else:
            return[]

    def getStartCoords(self):
        begin,end = self.getNodeTokenIterators()
        return coords(begin.deref().start[1], begin.deref().start[0])

    def getEndCoords(self):
        begin,end = self.getNodeTokenIterators()
        it = end.clone().predecr()
        return coords(it.deref().end[1], it.deref().end[0])



class ModuleTokenHandler(SimpleTokenHandler):
    pass

class ClassTokenHandler(SimpleTokenHandler):
    def rename(self, name):
        self.name = name
        # update the token
        begin,end = self.getNodeTokenIterators()
        it = begin.clone()
        it.preincr().deref().name = name
        #update the node
        self.fqn = self.parentScope.fqn+"."+name
        #update the typeinfo
        self.sourcenode.topnode.typeinfo.addType(self.fqn, self)
        self.parentScope.typeinfo.addType(name, self)
        self.sourcenode.setModified()


# methods which update the Function and keep tokens in sync
class FunctionTokenHandler(SimpleTokenHandler):

    def rename(self, name):
        self.sourcenode.tokenize()
        oldname = self.name
        self.name = name
        
        # update the token
        begin,end = self.getNodeTokenIterators()
        it = begin.clone()
        it.preincr().deref().name = name
        #update the node
        oldfqn = self.fqn
        newfqn = self.parentScope.fqn+"."+name
        self.fqn = newfqn
        #update the typeinfo                
        self.sourcenode.topnode.typeinfo.addType(self.fqn, self)
        self.parentScope.typeinfo.addType(name, self)
        self.sourcenode.setModified()


class GetattrTokenHandler(SimpleTokenHandler):
    def renameAttr(self, newname):
        begin,end = self.getNodeTokenIterators()
        it = begin.clone()
        for token in it:
            if token.name == self.attrname:
                token.name = newname
        self.attrname = newname
        self.sourcenode.setModified()

        
class NameTokenHandler(SimpleTokenHandler):
    def rename(self, newname):
        begin,end = self.getNodeTokenIterators()
        it = begin.clone()
        it.deref().name = newname
        oldname = self.name
        self.name = newname
        if self.parentScope.typeinfo.haskey(oldname):
            type = self.parentScope.typeinfo.getTypeOf(oldname)
            self.parentScope.typeinfo.addType(newname, type)
        self.sourcenode.setModified()
        

class ImportTokenHandler(SimpleTokenHandler):

    def rename(self, indexOfImport, newname):
        name = self.names[indexOfImport]
        oldfqn = name[0]
        oldname = fqn_rcar(oldfqn)
        self.names[indexOfImport] = (fqn_rcdr(oldfqn)+"."+newname, name[1])
        tbegin,tend = self._getTokensCorrespondingToImportName(indexOfImport)
        it = tend.clone().predecr()
        it.deref().name = newname
        self.sourcenode.setModified()
    
    # needs refactoring
    def _getTokensCorrespondingToImportName(self, indexOfImport):
        
        begin,end = self.getNodeTokenIterators()
        it = begin.clone().preincr()
        begintokenit = begin
        endtokenit = end
        commacount = 0
        while it != end:
            if commacount == indexOfImport:
                begintokenit = it.clone()
                break
            if it.deref().name == ",":
                commacount+=1
            it.preincr()

        def isAnImportToken(token):
            return token.name == "."or token.type == tokenize.NAME

        while it != end:
            if not isAnImportToken(it.deref()):
                endtokenit = it.clone()
            it.preincr()
        return begintokenit,endtokenit

class FromTokenHandler(SimpleTokenHandler):

    def getImportTokens(self):
        begin,end = self.getTokenIterators()
        it = begin.clone()
        def isImportToken(token):
            return token.type == tokenize.NAME and token.name == "import"
        while not isImportToken(it.deref()):
            it.preincr()
        return it,end

    def rename(self, indexOfImport, newname):
        name = self.names[indexOfImport]
        oldname = name[0]
        oldfqn = self.modname+"."+oldname
        self.names[indexOfImport] = (newname, name[1])
        nodetokenIterators = self.getImportTokens()
        tokit = self._getIndexOfTokenCorrespondingToImportName(nodetokenIterators, 
                                                                indexOfImport)
        tokit.deref().name = newname
        self.sourcenode.setModified()

    def _getIndexOfTokenCorrespondingToImportName(self, tokeniterators, indexOfImport):
        begin,end = tokeniterators
        it = begin.clone().preincr()
        commacount = 0
        while it != end:
            if commacount == indexOfImport:
                return it
            if it.deref().name == ",":
                commacount+=1
            it.preincr()
