#
# Typeinfo objects deduce the types of references within their scopes.
#   (they dynamically do call graph checking)
import compiler
import bike.parsing.extended_ast
from utils import fqn_car, fqn_cdr

class TypeNotSupportedException:
    def __init__(self,msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class TypeInfo:
    def __init__(self,parentScope=None):
        self.typehash = {}
        self.parentScope = parentScope
    def addType(self,name,node):
        self.typehash[name] = node
    def removeType(self,name):
        del self.typehash[name]
        
    def addReturnType(self,type):
        self.typehash['return'] = type

        
    def getTypeOf(self,name):
        try:
            return self._getTypeOf_Impl(name)
        except RuntimeError, ex:   # if recursive call exceeds maximum depth
            if str(ex) == "maximum recursion depth exceeded":
                #print "warning - averted an overflow"
                raise KeyError   # feign a lookup failure
            else:
                raise ex   # this will bugger up the stack trace. Don't know how to solve this

    def getTypeOf_doesntThrowException(self,name):
        try:
            return self._getTypeOf_Impl(name)
        except KeyError:
            return None
        except RuntimeError, ex:   # if recursive call exceeds maximum depth
            if str(ex) == "maximum recursion depth exceeded":
                #print "warning - averted an overflow2"
                return None
            else:
                raise ex   # this will bugger up the stack trace. Don't know how to solve this

    def _getTypeOf_Impl(self,name):
        typeof = None
        if "." in name:   # i.e. fqn
            typeof = self.typehash[fqn_car(name)].typeinfo.getTypeOf(fqn_cdr(name))
        else:
            typeof = self.typehash[name]
        if type(typeof) == type(lambda :1):  # is a closure
            #print "evaluating type"
            typeof = typeof()   # evaluate the closure
        return typeof

    def getReturnType(self):
        return self.getTypeOf_doesntThrowException('return')

    def haskey(self,name):
        return self.typehash.has_key(name)

    def __str__(self):
        mystr = ""
        for key in self.typehash.keys():
            mystr += key + ":" + str(self.getTypeOf_doesntThrowException(key)) + "\n"
        return mystr
        #return str(self.typehash)


    # returns None if can't deduce type of expression.
    def getTypeOfExpr(self,ast):
        if isinstance(ast,compiler.ast.Name):
            try:
                return self.getTypeOf(ast.name)
            except KeyError:
                if self.parentScope is not None:
                    return self.parentScope.typeinfo.getTypeOfExpr(ast)
        elif isinstance(ast,compiler.ast.Getattr):
            expr = self.getTypeOfExpr(ast.expr)
            if expr is not None:
                if isinstance(expr,bike.parsing.extended_ast.Instance):
                    return expr.getType().typeinfo.getTypeOf_doesntThrowException(ast.attrname)
                else:                    
                    return expr.typeinfo.getTypeOf_doesntThrowException(ast.attrname)
            return None
        else:
            raise TypeNotSupportedException, "Evaluation of "+str(ast)+" not supported"

# remove me!
class ClassTypeInfo(TypeInfo):
    def __init__(self,parentScope):
        TypeInfo.__init__(self,parentScope)


class ModuleTypeInfo(TypeInfo):
    def __init__(self,parentScope):
        TypeInfo.__init__(self,parentScope)
        self._modulesImportedWithStar = []

    def getTypeOf(self,name):
        type = TypeInfo.getTypeOf_doesntThrowException(self,name)
        if type is None:
            for modname in self._modulesImportedWithStar:
                module = self.parentScope.typeinfo.getTypeOf(modname)
                type = module.typeinfo.getTypeOf_doesntThrowException(name)
                if type is not None:
                    break
            if type is None:
                raise KeyError
        return type

    def addModuleImportedWithStar(self,modulename):
        self._modulesImportedWithStar.append(modulename)
