from bike.parsing.doublelinkedlist import region
import tokenize
from bike.parsing.tokenutils import createToken
import compiler
from bike.parsing.output import tokensToString
from bike.parsing.brmtransformer import parse


def extractMethod(sourceobj, startcoords, endcoords, newname):
    ExtractMethod().execute(sourceobj, startcoords, endcoords, newname)

def extractFunction(sourceobj, startcoords, endcoords, newname):
    ExtractFunction().execute(sourceobj, startcoords, endcoords, newname)


class ExtractMethod(object):
    def execute(self, sourceobj, startcoords, endcoords, newname):
        startcoords, endcoords = self.reverseCoordsIfWrongWayRound(startcoords, endcoords)

        tokens = sourceobj.getTokens()

        start_ptr, end_ptr = self._getTokenRegionToExtract(tokens, startcoords, endcoords)
        if isWhitespaceToken(start_ptr.clone().predecr().deref())or \
           start_ptr.clone().predecr().deref().name == ":":
            self._extractBlock(tokens, start_ptr, end_ptr, newname)
        else:
            self.extractExpression(tokens, start_ptr, end_ptr, newname)

        sourceobj.refreshTokens()
        sourceobj.setModified()

    def _getTokenRegionToExtract(self, tokens, startcoords, endcoords):
        start_ptr = findTokenWithCoords(region(tokens.begin(), tokens.end()), 
                                       startcoords.x, startcoords.y)
        # skip leading whitespace
        while isWhitespaceToken(start_ptr.deref()):
            start_ptr.preincr()

        end_ptr = findTokenWithCoords(region(start_ptr, tokens.end()), 
                                     endcoords.x, endcoords.y)
        # skip post selection whitespace
        while isWhitespaceToken(end_ptr.clone().predecr().deref()):
            end_ptr.predecr()
        return start_ptr, end_ptr


    def reverseCoordsIfWrongWayRound(self, startcoords, endcoords):
        if(startcoords.y > endcoords.y)or \
        (startcoords.y == endcoords.y and startcoords.x > endcoords.x):
            tmp = startcoords
            startcoords = endcoords
            endcoords = tmp
        return startcoords, endcoords

    def extractExpression(self, tokens, start_ptr, end_ptr, newname):

        fnregion, fnbodyptr = getSurroundingFunctionTokens(start_ptr)

        fnast = tokensToAST(fnregion.begin, fnregion.end)

        fnbodyptr.predecr() # reverse, so that this ptr doesnt get moved with the body extraction
        fnendptr = fnregion.end.clone().predecr()  # ditto with the end pointer

        argsptr, bodyptr = self.createFunctionAndReturnPtrToArgsAndBody(newname, tokens, 
                                                                        fnregion.end)

        tokens.insert(bodyptr, createToken(tokenize.NAME, "return"))
        bodyptr.preincr()
        tokens.moveRegion(start_ptr, end_ptr, bodyptr)
        newfnregion, bodyptr = getSurroundingFunctionTokens(bodyptr)
        newfnend = newfnregion.end.clone().predecr()
        tokens.insert(newfnend, createToken(tokenize.NEWLINE, ""))


        fnbodyptr.preincr()   # put the fn ptrs back to correct position
        fnendptr.preincr()

        callArgsPtr, endcallptr = self.insertFunctionCall(newname, tokens, end_ptr)

        fnast = tokensToAST(fnregion.begin, fnregion.end)
        beforecallast = tokensToAST(fnbodyptr, endcallptr)
        newfnast = tokensToAST(newfnregion.begin, newfnregion.end)

        self.writeFnArgs(tokens, fnast, argsptr, callArgsPtr, beforecallast, newfnast)


    def _extractBlock(self, tokens, start_ptr, end_ptr, newname):
        self._expandRegionToBalanceIndents(tokens, start_ptr, end_ptr)

        fnregion, fnbodyptr = getSurroundingFunctionTokens(start_ptr)

        fnast = tokensToAST(fnregion.begin, fnregion.end)

        fnbodyptr.predecr() # reverse, so that this ptr doesnt get moved with the body extraction
        fnendptr = fnregion.end.clone().predecr()  # ditto with the end pointer

        argsptr, bodyptr = self.createFunctionAndReturnPtrToArgsAndBody(newname, tokens, 
                                                                        fnregion.end)

        tokens.moveRegion(start_ptr, end_ptr, bodyptr)


        newfnregion, bodyptr = getSurroundingFunctionTokens(bodyptr)
        newfnend = newfnregion.end.clone().predecr()

        # if last extracted token wasnt newline or dedent, put in a newline
        prevtokentype = newfnend.clone().predecr().deref().type
        if(prevtokentype != tokenize.NEWLINE)and \
        (prevtokentype != tokenize.DEDENT):
            tokens.insert(newfnend, createToken(tokenize.NEWLINE, ""))
            newfnend.preincr()


        fncallptr = end_ptr.clone()

        fncallptr.predecr()
        callArgsPtr, endcallptr = self.insertFunctionCall(newname, tokens, end_ptr)
        fncallptr.preincr()

        fnbodyptr.preincr()
        fnendptr.preincr()


        beforecallast = tokensToAST(fnbodyptr, endcallptr)
        aftercallast = tokensToAST(endcallptr, fnendptr)
        newfnast = tokensToAST(newfnregion.begin, newfnregion.end)

        self.writeFnArgs(tokens, fnast, argsptr, callArgsPtr, beforecallast, newfnast)
        self.writeReturnParameters(tokens, newfnend, fncallptr, aftercallast, newfnast)

    def _expandRegionToBalanceIndents(self, tokens, start_ptr, end_ptr):
        requiredDedents = checkIfRegionIndentsAreBalanced(start_ptr, end_ptr)

        if requiredDedents:
            while requiredDedents:
                if end_ptr.deref().type == tokenize.DEDENT:
                    #print "found one"
                    requiredDedents-=1
                if end_ptr.deref().type == tokenize.INDENT:
                    raise "eerrr - didn't expect that!"
                end_ptr.preincr()
                # we have undoubtedly gobbled a newline, so put one in
            tokens.insert(end_ptr, createToken(tokenize.NEWLINE, ""))



    def writeFnArgs(self, tokens, fnast, argsptr, callArgsPtr, beforecallast, newfnast):
        temps = getAssignments(beforecallast)
        temps+=getArgumentsExceptSelf(fnast)

        newfnrefs = getReferences(newfnast)
        needcomma = 0
        for temp in temps:
            if temp in newfnrefs:
                if needcomma:
                    tokens.insert(callArgsPtr, createToken(tokenize.OP, ","))
                    callArgsPtr.preincr()
                tokens.insert(argsptr, createToken(tokenize.OP, ","))
                argsptr.preincr()
                tokens.insert(callArgsPtr, createToken(tokenize.NAME, temp))
                callArgsPtr.preincr()
                tokens.insert(argsptr, createToken(tokenize.NAME, temp))
                argsptr.preincr()
                needcomma = 1


    def writeReturnParameters(self, tokens, newfnend, fncallptr, aftercallast, newfnast):
        newfntemps = getAssignments(newfnast)
        aftercallrefs = getReferences(aftercallast)

        needreturn = 1
        haveretvals = 0
        for temp in newfntemps:
            if temp in aftercallrefs:
                haveretvals = 1
                if needreturn:
                    tokens.insert(newfnend, createToken(tokenize.NAME, "return"))
                    newfnend.preincr()
                    needreturn = 0
                    tokens.insert(newfnend, createToken(tokenize.NAME, temp))
                    newfnend.preincr()

                    tokens.insert(fncallptr, createToken(tokenize.NAME, temp))
                    fncallptr.preincr()
                else:
                    tokens.insert(newfnend, createToken(tokenize.OP, ","))
                    newfnend.preincr()
                    tokens.insert(newfnend, createToken(tokenize.NAME, temp))
                    newfnend.preincr()
                    tokens.insert(fncallptr, createToken(tokenize.OP, ","))
                    fncallptr.preincr()
                    tokens.insert(fncallptr, createToken(tokenize.NAME, temp))
                    fncallptr.preincr()
        if haveretvals:
            tokens.insert(fncallptr, createToken(tokenize.OP, "="))
            fncallptr.preincr()
            tokens.insert(newfnend, createToken(tokenize.NEWLINE, ""))
            newfnend.preincr()


    def createFunctionAndReturnPtrToArgsAndBody(self, name, tokens, it):
        it = it.clone()
        tokens.insert(it, createToken(tokenize.NL, ""))
        tokens.insert(it.preincr(), createToken(tokenize.NAME, "def"))
        tokens.insert(it.preincr(), createToken(tokenize.NAME, name))
        tokens.insert(it.preincr(), createToken(tokenize.OP, '('))
        tokens.insert(it.preincr(), createToken(tokenize.NAME, "self"))
        args = it.clone()
        tokens.insert(it.preincr(), createToken(tokenize.OP, ')'))
        tokens.insert(it.preincr(), createToken(tokenize.OP, ':'))
        tokens.insert(it.preincr(), createToken(tokenize.NEWLINE, ""))
        tokens.insert(it.preincr(), createToken(tokenize.INDENT, ""))
        body = it.clone()
        tokens.insert(it.preincr(), createToken(tokenize.DEDENT, ""))
        return args.preincr(), body.preincr()

    def insertFunctionCall(self, name, tokens, it):
        it = it.clone()
        tokens.insert(it, createToken(tokenize.NAME, "self"))
        tokens.insert(it.preincr(), createToken(tokenize.OP, '.'))
        tokens.insert(it.preincr(), createToken(tokenize.NAME, name))
        tokens.insert(it.preincr(), createToken(tokenize.OP, '('))
        args = it.clone()
        tokens.insert(it.preincr(), createToken(tokenize.OP, ')'))
        return args.preincr(), it.preincr()




class ExtractFunction(ExtractMethod):
    def createFunctionAndReturnPtrToArgsAndBody(self, name, tokens, it):
        it = it.clone()
        tokens.insert(it, createToken(tokenize.NL, ""))
        tokens.insert(it.preincr(), createToken(tokenize.NAME, "def"))
        tokens.insert(it.preincr(), createToken(tokenize.NAME, name))
        tokens.insert(it.preincr(), createToken(tokenize.OP, '('))
        args = it.clone()
        tokens.insert(it.preincr(), createToken(tokenize.OP, ')'))
        tokens.insert(it.preincr(), createToken(tokenize.OP, ':'))
        tokens.insert(it.preincr(), createToken(tokenize.NEWLINE, ""))
        tokens.insert(it.preincr(), createToken(tokenize.INDENT, ""))
        body = it.clone()
        tokens.insert(it.preincr(), createToken(tokenize.DEDENT, ""))
        return args.preincr(), body.preincr()

    def insertFunctionCall(self, name, tokens, it):
        it = it.clone()
        tokens.insert(it, createToken(tokenize.NAME, name))
        tokens.insert(it.preincr(), createToken(tokenize.OP, '('))
        args = it.clone()
        tokens.insert(it.preincr(), createToken(tokenize.OP, ')'))
        return args.preincr(), it.preincr()

    def writeFnArgs(self, tokens, fnast, argsptr, callArgsPtr, beforecallast, newfnast):
        temps = getAssignments(beforecallast)
        temps+=getArgumentsExceptSelf(fnast)

        newfnrefs = getReferences(newfnast)
        needcomma = 0
        for temp in temps:
            if temp in newfnrefs:
                if needcomma:
                    tokens.insert(argsptr, createToken(tokenize.OP, ","))
                    argsptr.preincr()
                    tokens.insert(callArgsPtr, createToken(tokenize.OP, ","))
                    callArgsPtr.preincr()
                tokens.insert(callArgsPtr, createToken(tokenize.NAME, temp))
                callArgsPtr.preincr()
                tokens.insert(argsptr, createToken(tokenize.NAME, temp))
                argsptr.preincr()
                needcomma = 1



def isWhitespaceToken(token):
    return token.type == tokenize.NEWLINE or \
           token.type == tokenize.DEDENT or \
           token.type == tokenize.NL or \
           token.type == tokenize.INDENT


def getAssignments(ast):
    class AssignVisitor:
        def __init__(self):
            self.assigns = []
        def visitAssign(self, node):
            for n in node.nodes:
                if isinstance(n, compiler.ast.Name)or \
                   isinstance(n, compiler.ast.AssName):
                    self.assigns.append(n.name)
                elif isinstance(n, compiler.ast.AssTuple):
                    for a in n.nodes:
                        self.assigns.append(a.name)
                else:
                    raise "don't know what type this is"+str(n)


    return compiler.walk(ast, AssignVisitor()).assigns

def getArgumentsExceptSelf(ast):
    class FunctionVisitor:
        def __init__(self):
            self.result = []
        def visitFunction(self, node):
            for n in node.argnames:
                if n != "self":
                    self.result.append(n)
    return compiler.walk(ast, FunctionVisitor()).result



def getReferences(ast):
    class NameVisitor:
        def __init__(self):
            self.result = []
        def visitName(self, node):
            self.result.append(node.name)
    return compiler.walk(ast, NameVisitor()).result


def tokensToAST(begin, end):
    src = tokensToString(begin, end)
    return parse(src+"\n")


def checkIfRegionIndentsAreBalanced(before, after):
    #print "checking balance",region(before,after)
    it = before.clone()
    requiredDedents = 0
    while it != after:
        if it.deref().type == tokenize.INDENT:
            requiredDedents+=1
        if it.deref().type == tokenize.DEDENT:
            requiredDedents-=1
        it.preincr()
    return requiredDedents

def createFunctionAndReturnPtrToArgsAndBody(name, tokens, it):
    it = it.clone()
    tokens.insert(it, createToken(tokenize.NL, ""))
    tokens.insert(it.preincr(), createToken(tokenize.NAME, "def"))
    tokens.insert(it.preincr(), createToken(tokenize.NAME, name))
    tokens.insert(it.preincr(), createToken(tokenize.OP, '('))
    tokens.insert(it.preincr(), createToken(tokenize.NAME, "self"))
    args = it.clone()
    tokens.insert(it.preincr(), createToken(tokenize.OP, ')'))
    tokens.insert(it.preincr(), createToken(tokenize.OP, ':'))
    tokens.insert(it.preincr(), createToken(tokenize.NEWLINE, ""))
    tokens.insert(it.preincr(), createToken(tokenize.INDENT, ""))
    body = it.clone()
    tokens.insert(it.preincr(), createToken(tokenize.DEDENT, ""))
    return args.preincr(), body.preincr()


def insertFunctionCall(name, tokens, it):
    it = it.clone()
    tokens.insert(it, createToken(tokenize.NAME, "self"))
    tokens.insert(it.preincr(), createToken(tokenize.OP, '.'))
    tokens.insert(it.preincr(), createToken(tokenize.NAME, name))
    tokens.insert(it.preincr(), createToken(tokenize.OP, '('))
    args = it.clone()
    tokens.insert(it.preincr(), createToken(tokenize.OP, ')'))
    return args.preincr(), it.preincr()




def findTokenWithCoords(region, x, y):
    it = region.begin.clone()
    while it != region.end:
        if it.deref().start[0] == y and \
           it.deref().start[1] >= x:
            break
        it.preincr()

    if it == region.end:
        raise "couldnt find the token with coords "+str(x)+","+str(y)+" in tokens "+str(region)
    return it

def getSurroundingFunctionTokens(it):
    startit = it.clone()
    while startit.deref().name != "def": startit.predecr()
    endit = startit.clone()
    indent = 0
    finish = 0

    while endit.deref().name != ":": endit.preincr()
    bodyptr = endit.preincr().clone()
    if endit.deref().type == tokenize.NEWLINE or \
       endit.deref().type == tokenize.COMMENT:  # is a multiline function

        while endit.deref().type != tokenize.INDENT: endit.preincr()
        bodyptr = endit.preincr().clone()
        indent = 1

        while indent > 0:
            if endit.deref().type == tokenize.INDENT: indent+=1
            if endit.deref().type == tokenize.DEDENT: indent-=1
            endit.preincr()
    else:
        while endit.deref().type != tokenize.NEWLINE: endit.preincr()
        endit.preincr()
    return region(startit, endit), bodyptr

