import compiler
from bike.parsing.utils import fqn_rcdr

# impl is delegated so can be reused by renamefunction
def renameClass(ast, fqn, newname, **opts):
    if opts.has_key('progressCallback') and opts['progressCallback'] is None:
        del opts['progressCallback']
        assert opts.has_key('progressCallback') == 0
        
    renameClass_impl(ast,fqn,newname,opts)

def renameClass_impl(ast, classfqn, newname, opts):
    theClass = ast.typeinfo.getTypeOf(classfqn)
    if opts.has_key('progressCallback'):
        opts['progressCallback']("Renaming Declaration", classfqn)    
    theClass.rename(newname)
    _renameClassReferences(ast,theClass,newname,opts)

def _renameClassReferences(ast,theclass,newname,opts):
    
    class Visitor:
        def visitModule(self, node):
            if opts.has_key('progressCallback'):
                opts['progressCallback']("Renaming References In", node.fqn)

            for child in node.getChildNodes():
                self.visit(child,node)

        def visitClass(self,node,scope):
            for child in node.getChildNodes():
                self.visit(child,node)
        
        def visitFunction(self,node,scope):
            for child in node.getChildNodes():
                self.visit(child,node)
        
        def visitGetattr(self,node,scope):
            type = None
            try:
                type = scope.typeinfo.getTypeOfExpr(node)
            except:    # TODO - don't seem to be able to import TypeNotSupportedException - this needs fixing
                pass
            if type == theclass:
                node.renameAttr(newname)
            for child in node.getChildNodes():
                self.visit(child,scope)

        def visitName(self,node,scope):
            #if node.name == "typeof":
                
                #print node.sourcenode.filename,node.parsetree
            if scope.typeinfo.getTypeOfExpr(node) == theclass:
                node.rename(newname)

        def visitImport(self,node,scope):
            for i in range(len(node.names)):
                name = node.names[i]            
                topnode = node.sourcenode.topnode
                if self._getImportedType(topnode,scope,name[0]) == theclass:
                    node.rename(i,newname)


        def visitFrom(self,node,scope):
            for i in range(len(node.names)):
                name = node.names[i]            
                topnode = node.sourcenode.topnode
                if self._getImportedType(topnode,scope,node.modname,name[0]) == theclass:
                    node.rename(i,newname)


        def _getImportedType(self,topnode,scope,modname,name=None):
            parent_package = fqn_rcdr(scope.fqn)
            if name == "*" or name == None:
                dot_name = ""
            else:
                dot_name = "."+name
            if topnode.typeinfo.getTypeOf_doesntThrowException(parent_package + "." +modname + dot_name):
                fqn = parent_package + "." +modname + dot_name
            elif topnode.typeinfo.getTypeOf_doesntThrowException(modname + dot_name):
                fqn = modname + dot_name
            else:
                return None  # can't find imported type
            return topnode.typeinfo.getTypeOf_doesntThrowException(fqn)

                
    visitor = compiler.walk(ast, Visitor())
    
    
