from bike.parsing.extended_ast import Top, Source, Package, Name, Class
from bike.parsing.utils import fqn_rcar, fqn_rcdr
import compiler

# Public Function
# returns (numberMethodsRenamed,numberMethodRefsRenamed)
# opts are 'progressCallback' and 'promptCallback'
def renameMethod(ast, methodfqn, nameAfter, **opts):
    if opts.has_key('progressCallback') and opts['progressCallback'] is None:
        del opts['progressCallback']
        assert opts.has_key('progressCallback') == 0

    assert isinstance(ast, Top)
    classname = fqn_rcdr(methodfqn)
    nameBefore = fqn_rcar(methodfqn)    
    theClass = ast.typeinfo.getTypeOf(classname)
    assert isinstance(theClass, Class)
    renameMethodClasses = _getAllClassesInHierarchy(theClass,ast)

    numMeths = 0
    for klass in renameMethodClasses:
        nMeths = _renameMethodDefinition(ast, klass.fqn, nameBefore, nameAfter, opts)
        numMeths+=nMeths

    classfqns = [klass.fqn for klass in renameMethodClasses]

    numRefs = _renameMethodReferences(ast, classfqns, nameBefore, nameAfter, opts)
                                                  
    return numMeths, numRefs

def _getAllClassesInHierarchy(klass,ast):
    rootclasses = _getRootClasses(klass)
    heirarchyClasses = []
    for klass in rootclasses:
        heirarchyClasses.append(klass)
        heirarchyClasses += _getAllSubClasses(klass,ast)
    return heirarchyClasses


def _getRootClasses(klass):
    if klass is None:  # i.e. dont have base class in our ast
        return None
    if len(klass.bases) == 0:  # i.e. is a root class
        return [klass]
    else:
        rootclasses = []
        for base in klass.bases:
            baseclass = klass.typeinfo.getTypeOfExpr(base)
            rootclass = _getRootClasses(baseclass)
            if rootclass is None:  # base class not in our ast
                rootclass = [klass]
            rootclasses += rootclass
        return rootclasses

def _getAllSubClasses(klass,ast,subclassmap=None):
    if subclassmap is None:
        subclassmap = SubclassMap(ast)
    subclasses = subclassmap.getSubClasses(klass)
    for subclass in subclasses:
        subclasses += _getAllSubClasses(subclass,ast,subclassmap)
    return subclasses


# TODO: optimize me! - move this to parsing module, invalidate on new load
class SubclassMap:
    def __init__(self,ast):
        self.map = {}
        map = self.map
        class ClassVisitor:
            def visitClass(self,node):
                for base in node.bases:
                    baseclass = node.typeinfo.getTypeOfExpr(base)
                    if baseclass is None:
                        return # do nothing
                    assert isinstance(baseclass,compiler.ast.Class)
                    if baseclass in map:
                        map[baseclass].append(node)
                    else:
                        map[baseclass] = [node]
        compiler.walk(ast, ClassVisitor())

    def getSubClasses(self,klass):
        try:
            return self.map[klass]
        except KeyError:
            return []


def _renameMethodDefinition(ast, classname, nameBefore, nameAfter, opts):
    fn = ast.typeinfo.getTypeOf_doesntThrowException(classname+"."+nameBefore)
    if fn is None:
        return 0  
    assert isinstance(fn, compiler.ast.Function)
    if opts.has_key('progressCallback'):
        opts['progressCallback']("Renaming Method", classname+"."+nameBefore)
    fn.rename(nameAfter)
    return 1

def _renameMethodReferences(ast, classnames, nameBefore, nameAfter, opts):
    promptCallback = None
    progressCallback = None
    if opts.has_key('progressCallback'):
        progressCallback = opts['progressCallback']
    if opts.has_key('promptCallback'):
        promptCallback = opts['promptCallback']

    class Visitor:

        def __init__(self, progressCallback):
            self.numRenames = 0
            self.progressCallback = progressCallback

        def visitSource(self, node):
            if self.progressCallback:
                self.progressCallback("Renaming Method References", node.filename)
            self.visit(node.module)

        def visitModule(self, node):
            for child in node.getChildNodes():
                self.visit(child,node)

        def visitClass(self,node,scope):
            for child in node.getChildNodes():
                self.visit(child,node)
        
        def visitFunction(self,node,scope):
            for child in node.getChildNodes():
                self.visit(child,node)
            
        def visitGetattr(self, node, scope):
            if node.attrname == nameBefore:
                try:
                    exprtype = scope.typeinfo.getTypeOfExpr(node.expr)
                    if exprtype.getType().fqn in classnames:
                        self._renameReference(node)
                except:
                    self._promptAndRename(node,scope)                
            self.visit(node.expr, scope)
            
        def _promptAndRename(self, node,scope):
            if promptCallback is not None:
                node.sourcenode.tokenize()
                if promptCallback(node, scope):
                    self._renameReference(node)


        def _renameReference(self, node):
            node.renameAttr(nameAfter)
            self.numRenames+=1

    visitor = compiler.walk(ast, Visitor(progressCallback))
    return visitor.numRenames



# pops the first element of the fully qualified name
# and returns the rest of the string, if the 
# element == nameToMatch
def _popNameIfEqual(fqname, nameToMatch):
    def nameUpToFirstDot(fqname):
        return fqname[: fqname.index(".")]
    def nameAfterFirstDot(fqname):
        return fqname[fqname.index(".")+1:]
    try:
        if nameUpToFirstDot(fqname) == nameToMatch:
            newfqname = nameAfterFirstDot(fqname)
        else: newfqname = fqname
    except ValueError:
        newfqname = fqname
    return newfqname

