#!/usr/bin/env python
import setpath
import unittest

from bike.refactor.extractMethod import *

from bike import testdata

from bike.parsing.tokenutils import getTokenList, getTokenVector
from bike.testutils import *
from bike.parsing.tokenhandler import coords
from bike.parsing import output

def assertTokensAreSame(t1begin, t1end, tokens):
    it = t1begin.clone()
    pos = 0
    while it != t1end:
        assert it.deref() == tokens[pos]
        it.preincr()
        pos+=1
    assert pos == len(tokens)


class TestExtractMethod(unittest.TestCase):

    def runTarget(self,src,begincoords,endcoords,newname):
        ast = createAST(src)
        extractMethod(ast,begincoords,endcoords,newname)
        return ast
    
    def test_extractsPass(self):
        ast = self.runTarget(testdata.extract_pass_before,
                             coords(8,4),coords(11,4),"newMethod")
        #print output.tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extract_pass_after))
        
    def test_extractsPassWhenFunctionAllOnOneLine(self):
        ast = self.runTarget(testdata.extractPassFromSingleLineFunction_before,
                             coords(24,3),coords(29,3),"newMethod")
        #print output.tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractPassFromSingleLineFunction_after))
        
    def test_extractsPassFromForLoop(self):
        ast = self.runTarget(testdata.extractPassFromForLoop_before,
                             coords(12,5),coords(16,5),"newMethod")
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractPassFromForLoop_after))

    def test_newMethodHasArgumentsForUsedTemporarys(self):
        #print getTokenVector(testdata.extractMethodWithTemporaryArgs_before)
        ast = self.runTarget(testdata.extractMethodWithTemporaryArgs_before,
                             coords(8,6),coords(21,7),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractMethodWithTemporaryArgs_after))
        
    def test_newMethodHasSingleArgument(self):
        ast = self.runTarget(testdata.extractMethodWithSingleTemporaryArg_before,
                             coords(8,5),coords(21,6),"newMethod")
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractMethodWithSingleTemporaryArg_after))


    def test_worksWhenAssignmentsToTuples(self):
        #print getTokenVector(testdata.extractMethodWithTupleAssignment_before)
        ast = self.runTarget(testdata.extractMethodWithTupleAssignment_before,
                             coords(8,5),coords(19,5),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractMethodWithTupleAssignment_after))


    def test_worksWhenUserSelectsABlockButDoesntSelectTheHangingDedent(self):
        #print getTokenVector(testdata.extractPassFromForLoop_before)
        ast = self.runTarget(testdata.extractPassFromForLoop_before,
                             coords(8,4),coords(16,5),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractPassFromForLoop_after2))


    def test_newMethodHasSingleReturnValue(self):
        #print getTokenVector(testdata.extractMethodWithSingleReturnValue_before)
        ast = self.runTarget(testdata.extractMethodWithSingleReturnValue_before,
                             coords(8,4),coords(34,4),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractMethodWithSingleReturnValue_after))

    def test_newMethodHasSingleReturnValue(self):
        #print getTokenVector(testdata.extractMethodWithSingleReturnValue_before)
        ast = self.runTarget(testdata.extractMethodWithSingleReturnValue_before,
                             coords(8,4),coords(34,4),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractMethodWithSingleReturnValue_after))

    def test_newMethodHasMultipleReturnValues(self):
        #print getTokenVector(testdata.extractMethodWithMultipleReturnValues_before)
        ast = self.runTarget(testdata.extractMethodWithMultipleReturnValues_before,
                             coords(8,4),coords(15,5),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractMethodWithMultipleReturnValues_after))

    def test_worksWhenMovingCodeJustAfterDedent(self):
        #print getTokenVector(testdata.extractCodeAfterDedent_before)
        ast = self.runTarget(testdata.extractCodeAfterDedent_before,
                             coords(8,6),coords(21,6),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractCodeAfterDedent_after))


    def test_skipsSelectedWhitespaceTokens(self):
        #print getTokenVector(testdata.extractCodeWithoutNewlines_before)
        ast = self.runTarget(testdata.extractCodeWithoutNewlines_before,
                             coords(16,5),coords(0,10),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractCodeWithoutNewlines_after))

    def test_extractsPassWhenSelectionCoordsAreReversed(self):
        ast = self.runTarget(testdata.extract_pass_before,
                             coords(11,4),coords(8,4),"newMethod")
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extract_pass_after))
        

    def test_extractsExpression(self):
        #print getTokenVector(testdata.extractPartOfExpression_before)
        ast = self.runTarget(testdata.extractPartOfExpression_before,
                             coords(16,5),coords(21,5),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractPartOfExpression_after))

    def test_extractsExpression2(self):
        #print getTokenVector(testdata.extractPartOfAssertExpression_before)
        ast = self.runTarget(testdata.extractPartOfAssertExpression_before,
                             coords(15,5),coords(32,5),"newMethod")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractPartOfAssertExpression_after))



        

class TestGetSurroundingFunctionTokens(unittest.TestCase):
    def test_getsTokensForSimpleCase(self):
        #ast = createAST(testdata.extract_pass_before)
        src = testdata.extract_pass_before
        tokens = getTokenList(src)
        it = tokens.begin()
        while it.deref().name != "pass":
            it.preincr()
        r,bodyptr = getSurroundingFunctionTokens(it)
        assertTokensAreSame(r.begin,r.end,getTokenVector(src)[6:-2])
        assert bodyptr.deref().name == "pass"
        
    def test_getsTokensForSingleLine(self):
        #ast = createAST(testdata.extract_pass_before)
        src = testdata.extractPassFromSingleLineFunction_before
        tokens = getTokenList(src)
        it = tokens.begin()
        while it.deref().name != "pass":
            it.preincr()
        r,bodyptr = getSurroundingFunctionTokens(it)
        assertTokensAreSame(r.begin,r.end,getTokenVector(src)[6:-2])
        assert bodyptr.deref().name == "pass"

class TestExtractFunction(unittest.TestCase):
    def runTarget(self,src,begincoords,endcoords,newname):
        ast = createAST(src)
        extractFunction(ast,begincoords,endcoords,newname)
        return ast

    def test_extractsFunction(self):
        #print getTokenVector(testdata.extractFunction_before)
        ast = self.runTarget(testdata.extractFunction_before,
                             coords(4,4),coords(13,5),"newFunction")
        #print tokensToString(ast.tokens.begin(),ast.tokens.end())
        self.assertEqual(ast.tokens,
                         getTokenList(testdata.extractFunction_after))


if __name__ == "__main__":
    unittest.main()
