#!/usr/bin/env python
import setpath
import unittest
from renameClass import *
from bike import testdata
from bike.testutils import *
from bike.parsing.extended_ast import Top
from bike.parsing.addtypeinfo import addtypeinfo
import compiler

class RenameClassTests:
    def runTarget(self, src, klassfqn, newname):
        # see concrete subclasses for implementation
        pass

    def testRenamesClassDcl(self):
        src = testdata.class_and_method
        ast = self.runTarget(src,"a","NewName")
        assert getASTNode(ast,"Class").name == "NewName"
        

    # i.e. a = TheClass()
    def testRenamesClassReferenceInInstanceCreation(self):
        src = testdata.RefCalledOnClassInstanceCreation
        ast = self.runTarget(src,"TheClass","NewName")
        assert getASTNode(ast,"CallFunc",1).node.name == "NewName"

    # i.e. if renaming TheClass, shouldnt rename a.b.c.TheClass
    def testDoesntRenameBugusClassReferenceOnEndOfGetattrNest(self):
        src = testdata.TheClassOnEndOfNestedExpression
        ast = self.runTarget(src,"TheClass","NewName")
        assert getASTNode(ast,"Getattr",0).attrname == "TheClass"

    def testRenamesClassRefUsedInExceptionRaise(self):
        src = testdata.RaiseClassException
        ast = self.runTarget(src,"TheClass","NewName")
        assert getASTNode(ast,"Name").name == "NewName"

    def testRenamesClassReferenceNameInInheritenceSpec(self):
        src = testdata.ClassDerivedFromOtherClassInSameModule
        ast = self.runTarget(src,"TheClass","NewName")
        assert getASTNode(ast,"Class",1).bases[0].name == "NewName"




class RenameClassTests_importsClass:
    def runTarget(self, src, klass, method, newname):
        """
        Must provide class in ast at a.b.bah.TheClass
        """

    def testRenamesClassReferenceInInstanceCreationWithFQN(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = self.runTarget(src,"a.b.bah.TheClass","NewName")
        assert getASTNode(ast,"CallFunc",0).node.attrname == "NewName"

    def testRenamesClassReferencesInInheritenceSpecs(self):        
        src = testdata.ClassWithNestedInheritanceSpec
        ast = self.runTarget(src,"a.b.bah.TheClass","NewName")
        assert getASTNode(ast,"Class",0).bases[0].attrname == "NewName"



class TestRenameClass(unittest.TestCase,RenameClassTests):
    def runTarget(self, src, classfqn, newname):
        ast = createSourceNodeAt(Top(),src,"foo")
        addtypeinfo(ast)
        renameClass(ast,"foo."+classfqn,newname)
        return ast
    
class TestRenameClassReferenceWithDirectoryStructure(unittest.TestCase,RenameClassTests_importsClass):
    def runTarget(self,src,classfqn,newname):
        ast = Top()
        ast = createSourceNodeAt(ast,src,"a.foo")
        ast = createSourceNodeAt(ast,testdata.TheClass,"a.b.bah")
        addtypeinfo(ast)
        renameClass(ast,classfqn,newname)
        return getASTNode(ast,"Module",1)
        
class TestRenameClassReferenceTwice(unittest.TestCase,RenameClassTests_importsClass):
    def runTarget(self,src,classfqn,newname):
        ast = Top()
        ast = createSourceNodeAt(ast,src,"a.foo")
        ast = createSourceNodeAt(ast,testdata.TheClass,"a.b.bah")
        addtypeinfo(ast)
        renameClass(ast,classfqn,"AnotherName")
        newfqn = fqn_rcdr(classfqn) + ".AnotherName"
        renameClass(ast,newfqn,newname)
        return getASTNode(ast,"Module",1)


if __name__ == "__main__":
    unittest.main()
