#!/usr/bin/env python
import setpath
import unittest
from renameFunction import *
from bike import testdata
from bike.testutils import *
from bike.parsing.extended_ast import Top
from bike.parsing.addtypeinfo import addtypeinfo
import compiler

class RenameFunctionTests:
    def runTarget(self, src, klassfqn, newname):
        # see concrete subclasses for implementation
        pass

    def testRenamesFunctionDcl(self):
        src = testdata.Function
        ast = self.runTarget(src,"theFunction","newName")
        assert getASTNode(ast,"Function").name == "newName"


class RenameFunctionTests_importsFunction:
    def runTarget(self, src, function, method, newname):
        """
        Must provide fn in ast at a.b.bah.theFunction, and then
        run renameFunction on it
        It should return the root node of the ast
        """

    def testRenamesImportedFunctionReference(self):
        src = testdata.FunctionRef
        ast = self.runTarget(src,"a.b.bah.theFunction","NewName")
        assert getASTNode(ast,"CallFunc").node.attrname == "NewName"
        

    def testRenamesFunctionReferenceImportedWithFromClause(self):
        src = testdata.FunctionRefImportedViaFrom
        ast = self.runTarget(src,"a.b.bah.theFunction","NewName")
        assert getASTNode(ast,"CallFunc").node.name == "NewName"


    def testRenamesFunctionRefInImportClause(self):
        src = testdata.FunctionRef
        ast = self.runTarget(src,"a.b.bah.theFunction","NewName")
        self.assertEqual(getASTNode(ast,"Import").names[0][0],
                         "b.bah.NewName")

    def testRenamesFunctionRefInImportClauseWhenItsTheSecondImportInTheStatement(self):
        src = testdata.ImportFunctionWithTwoImportsInSameStatement
        ast = self.runTarget(src,"a.b.bah.theFunction","NewName")
        self.assertEqual(getASTNode(ast,"Import").names[1][0],
                         "b.bah.NewName")

    
    def testRenamesFunctionRefInImportFromClause(self):
        src = testdata.FunctionRefImportedViaFrom
        ast = self.runTarget(src,"a.b.bah.theFunction","NewName")
        assert getASTNode(ast,"From").names[0][0] == "NewName"



class TestRenameFunction(unittest.TestCase,RenameFunctionTests):
    def runTarget(self, src, fqn, newname):
        ast = createSourceNodeAt(Top(),src,"foo")
        addtypeinfo(ast)
        renameFunction(ast,"foo."+fqn,newname)
        return ast


class TestRenameFunctionReferenceWithDirectoryStructure(unittest.TestCase,RenameFunctionTests_importsFunction):
    def runTarget(self,src,fqn,newname):
        ast = Top()
        ast = createSourceNodeAt(ast,src,"a.foo")
        ast = createSourceNodeAt(ast,testdata.Function,"a.b.bah")
        addtypeinfo(ast)
        renameFunction(ast,fqn,newname)
        return ast

# checks that renameFunction can be done twice on the same ast
class TestRenameFunctionReferenceTwice(unittest.TestCase,RenameFunctionTests_importsFunction):
    def runTarget(self,src,fqn,newname):
        ast = Top()
        ast = createSourceNodeAt(ast,src,"a.foo")
        ast = createSourceNodeAt(ast,testdata.Function,"a.b.bah")
        addtypeinfo(ast)
        renameFunction(ast,fqn,"AnotherName")
        newfqn = fqn_rcdr(fqn) + ".AnotherName"
        renameFunction(ast,newfqn,newname)
        return ast



if __name__ == "__main__":
    unittest.main()
