#!/usr/bin/env python
import unittest
import setpath

from bike import testdata
from bike.testutils import *
import bike
from bike.refactor.test_renameFunction import RenameFunctionTests, RenameFunctionTests_importsFunction
from bike.refactor.test_renameClass import RenameClassTests, RenameClassTests_importsClass
from bike.refactor.test_renameMethod import RenameMethodTests, RenameMethodTests_ImportsClass, RenameMethodReferenceTests, RenameMethodReferenceTests_ImportsClass, RenameMethodAfterPromptTests
from bike.refactor import test_extractMethod
import bikefacade
from bike import UndoStackEmptyException

class TestRenameMethod(RenameMethodTests, unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()

class TestRenameMethod_importsClass(RenameMethodTests_ImportsClass, 
                                    unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameMethod(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            ast = Top()
            ast.addSource(Source(tmpfile,tmpmodule,newsrc))
            addtypeinfo(ast)
            return ast
        finally:
            removePackageStructure()

class TestRenameMethodReferences(RenameMethodReferenceTests, 
                                 unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()

class TestRenameMethodReferences_importsClass(RenameMethodReferenceTests_ImportsClass, 
                                              unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameMethod(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            ast = Top()
            s = Source(tmpfile,tmpmodule,newsrc)
            ast.addSource(s)
            addtypeinfo(ast)
            return s.module
        finally:
            removePackageStructure()


class TestRenameMethodAfterPrompt(RenameMethodAfterPromptTests, 
                                  unittest.TestCase):
    def callback(self, filename, startcoords, endcoords):
        return 1

    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.setRenameMethodPromptCallback(self.callback)
            ctx.renameMethod(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()

class TestRenameFunction(RenameFunctionTests, unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameFunction(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()

class TestRenameFunction_importsFunction(RenameFunctionTests_importsFunction, 
                                   unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.Function)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameFunction(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            ast = parse(newsrc)
            return ast
        finally:
            removePackageStructure()

class TestRenameClass(RenameClassTests, unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.renameClass(tmpmodule+"."+fqn, newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()


class TestRenameClass_importsClass(RenameClassTests_importsClass, 
                                   unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            createPackageStructure(src, testdata.TheClass)
            ctx = bike.load(pkgstructureBasedir)
            ctx.renameClass(fqn, newname)
            ctx.save()
            newsrc = readFile(os.path.join(pkgstructureBasedir, "foo.py"))
            ast = parse(newsrc)
            return ast
        finally:
            removePackageStructure()

class TestRenameByCoordinates(RenameFunctionTests,
                              RenameClassTests,
                              RenameMethodTests,
                              unittest.TestCase):

    def getCoordinatesOfNode(self,ctx,fqn):
        node = ctx._getAST().typeinfo.getTypeOf(tmpmodule+"."+fqn)
        return node.getStartCoords().x,node.getStartCoords().y
    
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            coords = self.getCoordinatesOfNode(ctx,fqn)
            ctx.renameByCoordinates(os.path.abspath(tmpfile),coords[0],coords[1],newname)
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()

    def test_raisesExceptionIfCantFindNodeFromCoordinates(self):
        try:
            writeTmpTestFile("pass")
            ctx = bike.load(tmpfile)
            try:
                ctx.renameByCoordinates(os.path.abspath(tmpfile),1,1,"NewName")
                self.fail("shouldnt get to here")
            except bike.CouldntLocateASTNodeFromCoordinatesException:
                pass
        finally:
            deleteTmpTestFile()
        



class TestLocateNode(unittest.TestCase):
    def test_findsFunctioninNestedFile(self):
        src = testdata.Function[1:]
        createPackageStructure("pass", src)
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        assert ctx._locateNode(filename, 8, 1).fqn == "a.b.bah.theFunction"

    def test_returnsNoneIfCantFindNode(self):
        src = testdata.Function[1:]
        createPackageStructure("pass", src)
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        assert ctx._locateNode(filename, 15, 2) == None

    def test_findsClassinNestedFile(self):
        src = testdata.RaiseClassException[1:]
        createPackageStructure("pass", src)
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        assert ctx._locateNode(filename, 6, 1).fqn == "a.b.bah.TheClass"


class TestReloadSource(unittest.TestCase):
    def test_changesASTToReflectNewSource(self):
        createPackageStructure("pass", "pass")
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        createPackageStructure("pass", "class TheClass: pass")
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        ctx.load(filename)        
        removePackageStructure()
        assert ctx._locateNode(filename, 6,1).fqn == "a.b.bah.TheClass"

    def test_canRenameClassOnNewReloadedCode(self):
        createPackageStructure(testdata.functionContaingObjectCreationAndMethodReferenceWithImportOfNestedClass, "pass")
        ctx = bike.load(pkgstructureBasedir)
        removePackageStructure()
        createPackageStructure("pass", "class TheClass: pass")
        filename = os.path.join(os.path.abspath(pkgstructureBasedir), "b", "bah.py")
        ctx.load(filename)
        ctx.renameClass("a.b.bah.TheClass","NewName")
        removePackageStructure()
        assert getASTNode(ctx.ast,"CallFunc").node.attrname == "NewName"


class Test_deducePackageOfFile(unittest.TestCase):
    def test_returnsEmptyStringIfFileNotInPackage(self):
        try:
            # this doesnt have __init__.py file, so
            # isnt package
            os.makedirs("a")
            writeFile(os.path.join("a","foo.py"),"pass")
            pkg = bikefacade._deducePackageOfFile(os.path.join("a","foo.py"))
            assert pkg == ""
        finally:
            os.remove(os.path.join("a","foo.py"))
            os.removedirs(os.path.join("a"))

    def test_returnsNestedPackage(self):
        try:
            os.makedirs(os.path.join("a","b"))
            writeFile(os.path.join("a","__init__.py"),"# ")
            writeFile(os.path.join("a","b","__init__.py"),"# ")
            writeFile(os.path.join("a","b","foo.py"),"pass")
            pkg = bikefacade._deducePackageOfFile(os.path.join("a","b","foo.py"))
            assert pkg == "a.b"
        finally:
            os.remove(os.path.join("a","__init__.py"))
            os.remove(os.path.join("a","b","__init__.py"))
            os.remove(os.path.join("a","b","foo.py"))
            os.removedirs(os.path.join("a","b"))
        

class TestRenameMethodClientCallback(unittest.TestCase):

    class MockCallback(Mock):
        def callback(self,filename,startcoords,endcoords):
            return 1
    
    def test_callbackGetsPassedTheFilenameCorrectly(self):
        writeTmpTestFile(testdata.blankLineBetweenObjCreationAndMethodCall[1:])
        ctx = bike.load(tmpfile)
        deleteTmpTestFile()
        m = TestRenameMethodClientCallback.MockCallback()
        ctx.setRenameMethodPromptCallback(m.callback)
        ctx.renameMethod(tmpmodule+".TheClass.theMethod","NewName")
        assert m.mockCalledMethods["callback"][0].getParam(0) == os.path.abspath(tmpfile)

    def test_callbackGetsPassedTheCoordsCorrectly(self):
        writeTmpTestFile(testdata.blankLineBetweenObjCreationAndMethodCall[1:])
        ctx = bike.load(tmpfile)
        deleteTmpTestFile()
        m = TestRenameMethodClientCallback.MockCallback()
        ctx.setRenameMethodPromptCallback(m.callback)
        ctx.renameMethod(tmpmodule+".TheClass.theMethod","NewName")
        assert m.mockCalledMethods["callback"][0].getParam(1).x == 0
        assert m.mockCalledMethods["callback"][0].getParam(1).y == 7
        assert m.mockCalledMethods["callback"][0].getParam(2).x == 11
        assert m.mockCalledMethods["callback"][0].getParam(2).y == 7

class TestLoad(unittest.TestCase):
    def test_throwsExceptionIfTryToLoadANonPythonFile(self):
        try:
            writeFile("aoeuaoeuaoeu","hello")
            try:
                bike.load("aoeuaoeuaoeu")
                self.fail("should get to here")
            except bike.NotAPythonModuleOrPackageException:
                pass
        finally:
            os.remove("aoeuaoeuaoeu")

class TestExtractMethod(test_extractMethod.TestExtractMethod):
    def runTarget(self,src,begincoords,endcoords,newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.extractMethod(os.path.abspath(tmpfile),begincoords.x, begincoords.y,
                              endcoords.x, endcoords.y,newname)
            ctx.save()
            src = readTmpTestFile()
            return createAST(src)
        finally:
            deleteTmpTestFile()

class TestExtractFunction(test_extractMethod.TestExtractFunction):
    def runTarget(self,src,begincoords,endcoords,newname):
        try:
            writeTmpTestFile(src)
            ctx = bike.load(tmpfile)
            ctx.extractFunction(os.path.abspath(tmpfile),begincoords.x, begincoords.y,
                              endcoords.x, endcoords.y,newname)
            ctx.save()
            src = readTmpTestFile()
            return createAST(src)
        finally:
            deleteTmpTestFile()


class TestUndo(unittest.TestCase):
    def test_undoesTheTextOfASingleFile(self):
        try:
            writeTmpTestFile(testdata.two_classes_with_same_named_method)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+".a.foo","c")
            ctx.save()
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            assert getASTNode(ast, "Function", 0).name == "c"
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            self.assertEqual(newsrc,
                             testdata.two_classes_with_same_named_method)
        finally:
            deleteTmpTestFile()
        

    def test_undoesTwoConsecutiveRefactorings(self):
        try:
            writeTmpTestFile(testdata.two_classes_with_same_named_method)
            ctx = bike.load(tmpfile)
            ctx.renameMethod(tmpmodule+".a.foo","c")
            ctx.save()

            newsrc1 = readFile(tmpfile)
            ast = parse(newsrc1)
            assert getASTNode(ast, "Function", 0).name == "c"

            ctx.renameMethod(tmpmodule+".a.c","d")
            ctx.save()

            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            assert getASTNode(ast, "Function", 0).name == "d"

            # 1st undo
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            self.assertEqual(newsrc,
                             newsrc1)

            # 2nd undo
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            self.assertEqual(newsrc,
                             testdata.two_classes_with_same_named_method)
        finally:
            deleteTmpTestFile()
        

    def test_undoesTheTextOfAFileTwice(self):
        for i in range(3):
            try:
                writeTmpTestFile(testdata.two_classes_with_same_named_method)
                ctx = bike.load(tmpfile)
                ctx.renameMethod(tmpmodule+".a.foo","c")
                ctx.save()
            
                newsrc = readFile(tmpfile)
                ast = parse(newsrc)
                assert getASTNode(ast, "Function", 0).name == "c"

                ctx.undo()
                ctx.save()
                newsrc = readFile(tmpfile)
                self.assertEqual(newsrc,
                                 testdata.two_classes_with_same_named_method)
                raisedexception=0
                try:
                    ctx.undo()
                except UndoStackEmptyException:
                    raisedexception=1
                assert raisedexception
            finally:
                deleteTmpTestFile()

    def test_undoesManualModificationsToFiles(self):
        try:
            writeTmpTestFile(testdata.two_classes_with_same_named_method)
            origsrc = readFile(tmpfile)
            ctx = bike.load(tmpfile)

            deleteTmpTestFile()
            writeTmpTestFile(testdata.class_and_method)
            import os
            ctx.load(tmpfile)
            newsrc = readFile(tmpfile)
            assert newsrc != origsrc
            ctx.undo()
            ctx.save()
            newsrc = readFile(tmpfile)
            assert newsrc == origsrc
        finally:
            deleteTmpTestFile()

if __name__ == "__main__":
    unittest.main()
