#!/usr/bin/env python
import setpath
import unittest
from testutils import *
import testdata
import sys
from bike.parsing.extended_ast import Top 


class TestCreateSourceFileAt(unittest.TestCase):
    def test_putsTheSourceFileInPackage(self):
        top = Top([],[],[])
        createSourceNodeAt(top,"pass","a.b.c.foo");
        assert top.getPackage("a").getPackage("b").getPackage("c").getSource("foo")
        
    def test_canPutASourceFileInAnAlreadyExistingPackage(self):
        top = Top([],[],[])
        createSourceNodeAt(top,"pass","a.foo");
        createSourceNodeAt(top,"pass","a.bah");
        assert top.getPackage("a").getSource("foo")
        assert top.getPackage("a").getSource("bah")


class TestCetMatchingASTNode(unittest.TestCase):
    def test_returnsNodeThatExists(self):
        ast = createSourceNodeAt(Top(),
                                 testdata.MethodOverriddenOnDerivedClassInstance,
                                 "bah")
        addtypeinfo(ast)
        fnnode = getMatchingASTNode(ast,"Function",fqn="bah.DerivedDerivedClass.theMethod")
        assert fnnode.fqn == "bah.DerivedDerivedClass.theMethod"

        
    def test_returnsNoneIfNodeDoesntExist(self):
        ast = createSourceNodeAt(Top(),
                                 testdata.MethodOverriddenOnDerivedClassInstance,
                                 "bah")
        addtypeinfo(ast)
        fnnode = getMatchingASTNode(ast,"Function",fqn="bah.DerivedDerivedClass.nottherightname")
        assert fnnode == None
        


if __name__ == "__main__":
    unittest.main()
