import os
import os.path
from mock import Mock
from bike.parsing.extended_ast import Package, Source, PackageAlreadyExistsException, Top
from parsing.brmtransformer import parse
from parsing.addtypeinfo import addtypeinfo
from parsing.utils import fqn_rcar, fqn_rcdr

tmproot = "tmproot"
tmpfile = os.path.join(tmproot, "bicyclerepairman_tmp_testfile.py")
tmpmodule = "bicyclerepairman_tmp_testfile"


def writeFile(filename, src):
    f = open(filename, "w+")
    f.write(src)
    f.close()

def readFile(filename):
    f = open(filename)
    src = f.read()
    f.close()
    return src

def writeTmpTestFile(src):
    #try:
    os.makedirs(tmproot)
    #except OSError:
    #    pass
    writeFile(tmpfile, src)

def readTmpTestFile():
    return readFile(tmpfile)

def deleteTmpTestFile():
    os.remove(tmpfile)
    os.removedirs(tmproot)


pkgstructureBasedir = os.path.join(tmproot, "a")
pkgstructureChilddir = os.path.join(pkgstructureBasedir, "b")
pkgstructureFile1 = os.path.join(pkgstructureBasedir, "foo.py")
pkgstructureFile2 = os.path.join(pkgstructureChilddir, "bah.py")

def createPackageStructure(src1, src2):
    try: os.makedirs(pkgstructureChilddir)
    except: pass
    writeFile(os.path.join(pkgstructureBasedir, "__init__.py"), "#")
    writeFile(os.path.join(pkgstructureChilddir, "__init__.py"), "#")
    writeFile(pkgstructureFile1, src1)
    writeFile(pkgstructureFile2, src2)

def removePackageStructure():
    os.remove(os.path.join(pkgstructureBasedir, "__init__.py"))
    os.remove(os.path.join(pkgstructureChilddir, "__init__.py"))
    os.remove(pkgstructureFile1)
    os.remove(pkgstructureFile2)
    os.removedirs(pkgstructureChilddir)


def createAST(src):
    ast = Top()
    s = Source(tmpfile, tmpmodule, src)
    ast.addSource(s)
    return s

def createSourceNodeAt(top, src, fqn):
    node = top
    modname = fqn_rcar(fqn)
    packagefqn = fqn_rcdr(fqn)
    if packagefqn != "":
        for packagename in packagefqn.split("."):
            try:
                node.addPackage(Package(packagename, packagename))
            except PackageAlreadyExistsException:
                pass
            node = node.getPackage(packagename)
    node.addSource(Source(modname, modname, src))
    return top


# uses compiler visitor interface along with some dynamic magic
# to get a node out of the ast tree obtained from the src
# N.B. index refers to the number of nodes to count before returning.
# e.g. for second function, nodename=Function, index=1
def getASTNodeFromSrc(src, nodename, index = 0):
    from bike.parsing.brmtransformer import parse
    ast = Top()
    ast.addSource(Source("foo.py", "foo", src))
    addtypeinfo(ast)
    return getASTNode(ast, nodename, index)

def getASTNode(ast, nodename, index = 0):
    import compiler
    class FindNode:
        def __init__(self, nodename):
            self.nodename = nodename
            self.index = 0
        def __getattr__(self, name):
            if(name == "visit"+self.nodename):
                return self.visitNode
            else:
                assert name != "node"
                raise AttributeError, name
        def visitNode(self, node, scope = 0):
            if self.index == index:
                self.node = node
            self.index+=1
            for child in node.getChildNodes():
                self.visit(child)

    f = FindNode(nodename)
    compiler.walk(ast, f)
    return f.node



def getMatchingASTNode(ast, nodename, **predicates):
    import compiler
    class FindNode:
        def __init__(self, nodename):
            self.nodename = nodename
            self.node = None
        def __getattr__(self, name):
            if(name == "visit"+self.nodename):
                return self.visitNode
            else:
                assert name != "node"
                raise AttributeError, name

        def _nodeMatchesPredicate(self, node, name, value):
            return hasattr(node, name)and getattr(node, name) == value

        def visitNode(self, node):
            thisIsTheNode = 1
            for name, value in predicates.items():
                if not self._nodeMatchesPredicate(node, name, value):
                    thisIsTheNode = 0
            if thisIsTheNode:
                self.node = node
            else:
                for child in node.getChildNodes():
                    self.visit(child)

    f = FindNode(nodename)
    compiler.walk(ast, f)
    return f.node
