# bicycle repair man idle extension
import bike
import bike.parsing.load
import os
from Tkinter import*
import tkFileDialog
import tkMessageBox
import tkSimpleDialog
import sys
import string
try:
    from idlelib.PathBrowser import*
    from idlelib.WindowList import ListedToplevel
    from idlelib.TreeWidget import TreeNode, TreeItem, ScrolledCanvas
    from idlelib import EditorWindow
    from idlelib.PyShell import PyShell
except ImportError:
    from PathBrowser import*
    from WindowList import ListedToplevel
    from TreeWidget import TreeNode, TreeItem, ScrolledCanvas
    import EditorWindow
    from PyShell import PyShell


brmctx = None
shellwin = None
loadingFiles = 0


class BicycleRepairMan_Idle:
    menudefs = [
    ('refactor', [
    ('_Import Package Into Bicycle Repair Man', '<<brm-load>>'), 
            ('Rename _Method', '<<brm-rename-method>>'), 
            ('Rename _Class', '<<brm-rename-class>>'), 
            ('Rename _Function', '<<brm-rename-function>>'), 
            ('_Extract Method', '<<brm-extract-method>>'), 
            ('E_xtract Function', '<<brm-extract-function>>'), 
            ('_Undo', '<<brm-undo>>'), 
         ])
    ]

    keydefs = {
    '<<brm-load>>':[], 
        '<<brm-rename-class>>':[], 
        '<<brm-rename-method>>':[], 
        '<<brm-rename-function>>':[], 
        '<<brm-extract-method>>':[], 
        '<<brm-extract-function>>':[], 
        '<<brm-undo>>':[], 
     }

    def __init__(self, editwin):
        self.editwin = editwin
        if not isinstance(editwin, PyShell):
            # sly'ly add the refactor menu to the window
            name, label = ("refactor", "_Refactor")
            underline, label = EditorWindow.prepstr(label)
            mbar = editwin.menubar
            editwin.menudict[name] = menu = Menu(mbar, name = name)
            mbar.add_cascade(label = label, menu = menu, underline = underline)
            self.orig_saved_change_hook = editwin.undo.saved_change_hook
            self.editwin.undo.set_saved_change_hook(self.saved_change_hook)

            # Initialize Bicyclerepairman and import the code
            path = self.editwin.io.filename
            if path is not None:
                global brmctx
                if not brmctx:
                    brmctx = bike.load(path, 
                                       ProgressWriter().printFilename)
                else:
                    brmctx.load(path)
        else:
            global shellwin
            shellwin = editwin

    def brm_load_event(self, event):
        try:
            # need to store this for closing later
            self.browser = BRMPathBrowser(self.editwin.flist, self.editwin.io.filename, self)
        except:
            self._handleUnexpectedException()

    def saved_change_hook(self):
        try:
            if self.editwin.get_saved():
                if not loadingFiles:
                    global brmctx
                    path = self.editwin.io.filename
                    if not brmctx:
                        brmctx = bike.load(path, 
                                           ProgressWriter().printFilename)
                    else:
                        brmctx.load(path)
            self.orig_saved_change_hook()
        except:
            self._handleUnexpectedException()

    def brm_rename_class_event(self, event):
        try:
            self.renameItemByCoordinates("Class")
        except:
            self._handleUnexpectedException()

    def brm_rename_method_event(self, event):
        try:
            self.renameItemByCoordinates("Method")
        except:
            self._handleUnexpectedException()

    def brm_rename_function_event(self, event):
        try:
            self.renameItemByCoordinates("Function")
        except:
            self._handleUnexpectedException()

    def brm_extract_method_event(self, event):
        try:
            if not self.confirm_all_buffers_saved():
                return
            filename, newname, beginline, begincolumn, endline, endcolumn = self._getExtractionInformation("Method")
            brmctx.extractMethod(filename, int(begincolumn), int(beginline), 
                                 int(endcolumn), int(endline), newname)
            savedfiles = brmctx.save()
            self.refreshWindows(savedfiles, beginline)
        except:
            self._handleUnexpectedException()

    def brm_extract_function_event(self, event):
        try:
            if not self.confirm_all_buffers_saved():
                return
            filename, newname, beginline, begincolumn, endline, endcolumn = self._getExtractionInformation("Function")

            brmctx.extractFunction(filename, int(begincolumn), int(beginline), 
                                 int(endcolumn), int(endline), newname)
            savedfiles = brmctx.save()
            self.refreshWindows(savedfiles, beginline)
        except:
            self._handleUnexpectedException()


    def brm_undo_event(self, event):
        try:
            line, column = string.split(self.editwin.text.index("insert"), '.')
            brmctx.undo()
            savedfiles = brmctx.save()
            self.refreshWindows(savedfiles, line)
        except:
            self._handleUnexpectedException()

    def _handleUnexpectedException(self):
        import traceback
        traceback.print_exc()
        self.errorbox("Caught Exception", "Caught Exception "+str(sys.exc_info()[0]))



    def _getExtractionInformation(self, extracttype):
        self.checkCodeImportedIntoBRM()
        if self.editwin.text.index("sel.first") == "":
            self.errorbox("Code not highlighted", "Highlight the region of code you want to extract and try again")
            return
        filename = os.path.normpath(self.editwin.io.filename)
        newname = tkSimpleDialog.askstring("Extract Method ", 
                                           "New "+extracttype+" Name:", 
                                           parent = self.editwin.text)
        beginline, begincolumn = string.split(self.editwin.text.index("sel.first"), '.')
        endline, endcolumn = string.split(self.editwin.text.index("sel.last"), '.')
        return filename, newname, beginline, begincolumn, endline, endcolumn

    def renameMethodPromptCallback(self, filename, startcoords, endcoords):
        editwin = self.editwin.flist.open(filename)
        originaltop = self.editwin.getwindowlines()[0]

        # select the method call and position the window
        editwin.text.tag_remove("sel", "1.0", "end")
        editwin.text.tag_add("sel", str(startcoords.y)+"."+str(startcoords.x), 
                                  str(endcoords.y)+"."+str(endcoords.x))

        line, column = string.split(editwin.text.index("sel.first"), '.')
        editwin.text.yview(str(int(line)-2)+".0")


        d = NoFocusDialog("Rename Method?", 
                     "Cannot deduce the type of highlighted object reference.\nRename this method call?", 
                     parent = editwin.text)

        # put the window back where it was
        self.editwin.text.yview(float(originaltop))
        return d.answer



    def checkCodeImportedIntoBRM(self):
        if brmctx == None:
            answer = tkMessageBox.askyesno("code not imported", "Your code isnt imported into Bicycle Repair Man. Do you want to import *just* this module?", master = self.editwin.text)
            if answer == 1:
                self.load(self.editwin.io.filename)
                return 1
            else:
                self.errorbox("import package", "Import a python package using the import menu option")
                return 0
            if brmctx == None:
                print "brmctx is still None - what's going on?"
                return 0   # well, we tried!
        else:
            return 1


    def renameItemByCoordinates(self, renameType):
        if not self.confirm_all_buffers_saved():
            return
        if self.editwin.text.index("sel.first") == "":
            self.errorbox("Name not highlighted", "Double click the name of the "+renameType+" declaration you want to rename (to highlight it) and try again")
            return

        if not self.checkCodeImportedIntoBRM():
            return

        brmctx.setRenameMethodPromptCallback(self.renameMethodPromptCallback)
        line, column = string.split(self.editwin.text.index("sel.first"), '.')
        print "line", line, "column", column
        filename = os.path.normpath(self.editwin.io.filename)
        newname = tkSimpleDialog.askstring("Rename "+renameType, 
                                           "Rename "+renameType+" To:", 
                                           parent = self.editwin.text)
        if newname is None: # cancel clicked
            return
        brmctx.renameByCoordinates(filename, int(column), int(line), newname)
        savedfiles = brmctx.save()
        self.refreshWindows(savedfiles, line)



    def refreshWindows(self, savedfiles, line):
        # refresh editor windows
        oldtop = self.editwin.getwindowlines()[0]

        global loadingFiles
        loadingFiles = 1
        for sf in savedfiles:
            normsf = os.path.normcase(sf)
            if normsf in self.editwin.flist.dict:
                print "reloading file"
                editwin = self.editwin.flist.dict[normsf]
                editwin.io.loadfile(sf)
        loadingFiles = 0

        self.editwin.text.mark_set("insert", float(line))
        self.editwin.text.yview(float(oldtop))



    def confirm_all_buffers_saved(self):
        filelist = self.editwin.flist.dict.keys()
        for f in filelist:
            #editwin = self.editwin.flist.open(f)
            editwin = self.editwin.flist.dict[f]
            if self.confirm_buffer_is_saved(editwin) == 0:
                return 0
        return 1


    def confirm_buffer_is_saved(self, editwin):
        if not editwin.get_saved():
            name = (editwin.short_title()or
            editwin.long_title()or
            "Untitled")
            self.errorbox("Not saved", 
                          "The buffer for %s is not saved.\n"%name+
            "Please save it first!")
            editwin.text.focus_set()
            return 0
        return 1

    def errorbox(self, title, message):
        tkMessageBox.showerror(title, message, master = self.editwin.text)
        self.editwin.text.focus_set()

    def load(self, path):
        global brmctx
        if hasattr(self, "browser")and self.browser:
            self.browser.close()
            del self.browser
        if not brmctx:
            brmctx = bike.load(path, ProgressWriter().printFilename)
        else:
            brmctx.load(path)
        tkMessageBox.showinfo("import successful", "code imported successfully", master = self.editwin.text)



class ProgressWriter:
    def __init__(self):
        #self.io = io
        self.io = sys.stdout

    def _write(self, txt):
        self.io.write(txt)
        self.io.flush()

    def printDot(self, action, filename):
        import sys
        if not hasattr(self, "currentAction"):
            self._write(action)
        elif self.currentAction != action:
            self._write("\n"+action)
        self.currentAction = action
        self._write(".")

    def printFilename(self, action, filename):
        import sys
        self._write("BRMMessage: "+action+" "+filename+"\n")


class NoFocusDialog(tkSimpleDialog._QueryDialog):
    def __init__(self, title, prompt, 
                 initialvalue = None, 
                 minvalue = None, maxvalue = None, 
                 parent = None):
        self.answer = 0

        if not parent:
            import Tkinter
            parent = Tkinter._default_root

        self.prompt = prompt
        self.minvalue = minvalue
        self.maxvalue = maxvalue

        self.initialvalue = initialvalue

        Toplevel.__init__(self, parent)

        if title:
            self.title(title)

        self.parent = parent

        self.result = None

        body = Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx = 5, pady = 5)

        self.buttonbox()

        self.grab_set()

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        if self.parent is not None:
            self.geometry("+%d+%d"%(parent.winfo_rootx()+50, 
                                      parent.winfo_rooty()+50))
        self.wait_window(self)


    def getresult(self):
        self.answer = 1

    def body(self, master):
        w = Label(master, text = self.prompt, justify = LEFT)
        w.grid(row = 0, padx = 5, sticky = W)

    def buttonbox(self):
        box = Frame(self)

        w = Button(box, text = "Yes", width = 10, command = self.ok, default = ACTIVE)
        w.pack(side = LEFT, padx = 5, pady = 5)
        w = Button(box, text = "No", width = 10, command = self.cancel)
        w.pack(side = LEFT, padx = 5, pady = 5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)
        box.pack()




class BRMPathBrowser(PathBrowser):

    def __init__(self, flist, editwin_filename, eventhandler):
        self.eventhandler = eventhandler
        self.editwin_filename = editwin_filename
        self.init(flist)

    def init(self, flist):
        self.flist = flist
        # create top
        self.top = top = ListedToplevel(flist.root)
        top.protocol("WM_DELETE_WINDOW", self.close)
        top.bind("<Escape>", self.close)
        self.settitle()
        top.focus_set()
        self.topbox = topbox = Frame(top)
        l = Label(topbox, 
                  text = "Double click a package or module to import it", 
                  justify = CENTER)
        l.grid(row = 0, column = 0, columnspan = 2)
        # create scrolled canvas
        sc = ScrolledCanvas(topbox, bg = "white", highlightthickness = 0, takefocus = 1)
        sc.frame.grid(row = 1, column = 0, columnspan = 2, sticky = "we")
        topbox.pack()
        item = self.rootnode()
        self.node = node = TreeNode(sc.canvas, None, item)
        node.update()
        node.expand()

    def rootnode(self):
        dir = bike.parsing.load.getRootDirectory(self.editwin_filename)
        return BRMPathBrowserTreeItem(self.eventhandler, [dir])


class BRMPathBrowserTreeItem(PathBrowserTreeItem):

    def __init__(self, eventhandler, extraPaths):
        self.extraPaths = extraPaths
        PathBrowserTreeItem.__init__(self)
        self.eventhandler = eventhandler

    def GetSubList(self):
        sublist = []
        for dir in self.extraPaths:
            item = BRMToplevelDirBrowserTreeItem(dir, self.eventhandler)
            sublist.append(item)

        for dir in sys.path:
            item = BRMToplevelDirBrowserTreeItem(dir, self.eventhandler)
            sublist.append(item)
        return sublist


class BRMToplevelDirBrowserTreeItem(DirBrowserTreeItem):

    def __init__(self, dir, eventhandler):
        DirBrowserTreeItem.__init__(self, dir)
        self.eventhandler = eventhandler


    def GetSubList(self):
        try:
            names = os.listdir(self.dir or os.curdir)
        except os.error:
            return[]
        directories = []
        for name in names:
            file = os.path.join(self.dir, name)
            if os.path.isdir(file):
                directories.append((file, name))
        directories.sort()
        sublist = []
        for file, name in directories:
            item = BRMDirBrowserTreeItem(file, name, self.eventhandler)
            sublist.append(item)
        for nn, name in self.listmodules(names):
            item = BRMModuleBrowserTreeItem(os.path.join(self.dir, name), 
                                            self.eventhandler)
            sublist.append(item)
        return sublist

    def GetText(self):
        return self.dir

    def OnDoubleClick(self):
        self.eventhandler.load(self.dir)


class BRMDirBrowserTreeItem(BRMToplevelDirBrowserTreeItem):
    def __init__(self, dir, name, eventhandler):
        self.dir = dir
        self.name = name
        self.eventhandler = eventhandler

    def GetText(self):
        return self.name


class BRMModuleBrowserTreeItem(ModuleBrowserTreeItem):
    def __init__(self, file, eventhandler):
        ModuleBrowserTreeItem.__init__(self, file)
        self.eventhandler = eventhandler

    def OnDoubleClick(self):
        self.eventhandler.load(self.file)

    def IsExpandable(self):
        return 0
