;
;   XEmacs support for bicycle repair man
;   -------------------------------------
;   Author: Phil Dawes (pdawes@users.sourceforge.net)
;
;   Eval this buffer, then do 'M-x brm-load' to load a python package
;   structure into bicycle repair man. Then do 'M-x brm-rename' to
;   rename a class/function/method at point.
;
;   If you're debugging, check the 'Bicycle Repair Man' buffer for trace

(require 'python-mode)

(defvar brm-menu nil "Menu for Python Mode.  This menu will get
created automatically if you have the `easymenu' package.  Note that
the latest X/Emacs releases contain this package.")


(and (py-safe (require 'easymenu) t)
	 (easy-menu-define
      brm-menu py-mode-map "Bicycle Repair Man "
      '("Refactor"
				   ["Import Code Into BRM" brm-load t]
		           "---"
				   ["Rename-Method" brm-rename t]
				   ["Rename-Function" brm-rename t]
				   ["Rename-Class" brm-rename t]
		           "---"
				   ["Extract-Method" brm-extract-method t]
				   ["Extract-Function" brm-extract-function t]
		           "---"
				   ["Undo Last Refactoring" brm-undo t]

	)))

(if brm-menu
	(add-hook 'python-mode-hook (lambda () (easy-menu-add brm-menu))))


;(add-hook 'python-mode-hook (lambda () (brm-reload-file)))



(defun brm-start ()
"starts the bicycle repair man process"
  (make-comint "Bicycle Repair Man" "bikeemacs.py")
  (set-process-filter (get-process "Bicycle Repair Man") `brm-process-filter) 
)

(defun brm-kill ()
"kills the current bicycle repair man process"
  (interactive)
  (if (get-process "Bicycle Repair Man")
	  (kill-buffer (process-buffer (get-process "Bicycle Repair Man"))))
)



(defun brm-load (path)
"Loads the python file or directory (package) structure into bicycle repair man"
  (interactive "fPython file or directory: ")
  (if (not (get-process "Bicycle Repair Man")) (brm-start))
  (process-send-string "Bicycle Repair Man" (concat "load " path "\n"))
)

(defun brm-rename (newname)
"Renames the class, function or method at point"
  (interactive "sNew name: ")
  (save-some-buffers)
  (process-send-string "Bicycle Repair Man" 
					   (format "rename %s %d %d %s\n" 
							   (buffer-file-name)
							   (current-column)
							   (count-lines 1 (point))
							   newname)))

(defun brm-extract-method (methodname)
"Extracts the selected region into a new method"
  (interactive "sMethod name: ")
  (save-some-buffers)
  (exchange-point-and-mark)  
  (save-excursion
	(setq beginx (current-column))
	(setq beginy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq beginy (+ beginy 1)))  ; get round 'if col is 0 then line is 1 less'
	(exchange-point-and-mark)  
	(setq endx (current-column))
	(setq endy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq endy (+ endy 1)))  ; get round 'if col is 0 then line is 1 less'

	(process-send-string "Bicycle Repair Man" 
						 (format "extractMethod %s %d %d %d %d %s\n" 
								 (buffer-file-name)
								 beginx beginy endx endy methodname))))


(defun brm-extract-function (functionname)
"Extracts the selected region into a new function"
  (interactive "sFunction name: ")
  (save-some-buffers)
  (exchange-point-and-mark)  
  (save-excursion
	(setq beginx (current-column))
	(setq beginy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq beginy (+ beginy 1)))  ; get round 'if col is 0 then line is 1 less'
	(exchange-point-and-mark)  
	(setq endx (current-column))
	(setq endy (count-lines 1 (point)))
	(if (= (current-column) 0) (setq endy (+ endy 1)))  ; get round 'if col is 0 then line is 1 less'

	(process-send-string "Bicycle Repair Man" 
						 (format "extractFunction %s %d %d %d %d %s\n" 
								 (buffer-file-name)
								 beginx beginy endx endy functionname))))


(defun brm-undo ()
"Undoes the last refactoring. WARNING: this is dangerous if \
the user has modified files since the last refactoring"
  (interactive)
  (process-send-string "Bicycle Repair Man" 
					   (format "undo\n")))



(defun brm-get-type ()
"Returns a string representation of the type of the python construct \
at point"
  (interactive)
  (process-send-string "Bicycle Repair Man" 
					   (format "gettype %s %d %d\n" 
							   (buffer-file-name)
							   (current-column)
							   (count-lines 1 (point))))
  (accept-process-output (get-process "Bicycle Repair Man"))
  (setq myresult brm-result)
  (setq brm-result nil)
  myresult
)

(defun brm-test-get-type ()
  (interactive)
  (setq res (brm-get-type))
  (if res (message res) (message "Type wasnt recognised"))
)

(defvar brm-result nil)

(defun brm-process-filter (proc string)
"Internal filter for the bicycle repair man process"
  (save-excursion
	(set-buffer (process-buffer (get-process "Bicycle Repair Man")))
	(insert string))
  
  (if (string-match "BRMError: \\(.*\\)$" string)
	  (error (match-string 1 string)))
  (if (string-match "BRMMessage: \\(.*\\)$" string)
	  (message (match-string 1 string)))
  (if (string-match "BRMUpdated: \\(.*\\)$" string)
	  (brm-revert-file-buffer (match-string 1 string)))
  (if (string-match "BRMResult: \\(.*\\)$" string)
	  (setq brm-result (match-string 1 string)))
  (if (string-match "BRMPromptForRename: \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\)$" string)
	  (if (brm-prompt-for-rename (match-string 1 string)
								 (string-to-number (match-string 2 string))
								 (string-to-number (match-string 3 string))
								 (string-to-number (match-string 4 string))
								 (string-to-number (match-string 5 string))
								 "Couldn't deduce object type - rename this method reference? ")
		  (process-send-string "Bicycle Repair Man" "y\n")
		(process-send-string "Bicycle Repair Man" "n\n")))
  )	

(defun brm-prompt-for-rename (filename startx starty endx endy message)
"Internal function - goes to file and prompts user, asking if need to rename"
  (setq mybuf (current-buffer))
  (setq ans (save-excursion
			  (find-file filename)
			  (goto-line starty)
			  (move-to-column startx)
			  (set-mark (+ (point) (- endx startx)))
			  (activate-region)
			  (y-or-n-p message)))
  (switch-to-buffer mybuf)
  ans
  )

;(if (brm-prompt-for-rename "Couldn't deduce object type - rename this method reference? ")

;(

;(y-or-n-p "this is a prompt?")

;(brm-prompt-for-rename "/home/pld/src/bicyclerepair/bike/ui/scrap.py" "5" "3" "Couldn't deduce object type - rename this method reference? ")

;(progn (if (brm-prompt-for-rename "/home/pld/src/bicyclerepair/bike/ui/scrap.py" 5 3 "Couldn't deduce object type - rename this method reference? ")
;		   (message "Bicycle Repair Man y")
;		 (message "Bicycle Repair Man n")))



(defvar brm-updating-buffer nil "Internal variable containing name of file to be updated, so that it doesnt get reloaded into brm")


(defun brm-revert-file-buffer (filename)
"reverts the buffer currently looking at filename, if it exists"
  (setq mybuf (find-buffer-visiting filename))
  (if mybuf
	  (progn (set-buffer mybuf)
			 (setq brm-updating-buffer filename)
			 (revert-buffer nil 1)
			 (setq brm-updating-buffer nil)
			 )
	(message "Couldnt find buffer")
	)
  )


;(setq brm-updating-buffer "/home/pld/src/bicyclerepair/ide-integration/bike.el")
;(equal brm-updating-buffer (buffer-file-name))
;(message (buffer-file-name))
;(message brm-updating-buffer)

(defun brm-reload-file ()
  (if (and (string-match ".py$" (buffer-file-name)) 
		   (not (equal brm-updating-buffer (buffer-file-name))))
	  (progn
	  (if (not (get-process "Bicycle Repair Man")) (brm-start))
	  (process-send-string "Bicycle Repair Man"		   
						   (format "load %s\n" (buffer-file-name)))
	  )
	)
)


(add-hook 'after-save-hook 'brm-reload-file)
(add-hook 'find-file-hooks 'brm-reload-file)


